/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file data.hh
	\brief Functions for reading data files
*/




#ifndef __DATA_HH__
#define __DATA_HH__

#ifdef __cplusplus




#include <boost/regex.hpp>
#include <boost/filesystem.hpp>

#include <boost/shared_array.hpp>
#include <boost/assert.hpp>

#include <algorithm>
#include <string>
#include <vector>
#include <stdexcept>
#include <tr1/functional>

#include <stdio.h>
#include <string.h>

#include <ply.hpp>




namespace {




//============================================================================
//    ReadPLYFileHelper class
//============================================================================


template< typename t_Type >
struct ReadPLYFileHelper {

	inline ReadPLYFileHelper();


	inline void Read(
		unsigned int& dimension,
		std::vector< boost::shared_array< t_Type > >& pPoints,
		boost::filesystem::path const& path
	);


private:

	std::tr1::function< void ( ply::float32 ) > DefineCallbacks( std::string const& elementName, std::string const& propertyName );

	void VertexCallback( unsigned int index, ply::float32 const value );


	ply::ply_parser m_parser;

	unsigned int m_index;
	boost::shared_array< t_Type > m_pPoint;

	unsigned int m_dimension;
	std::vector< boost::shared_array< t_Type > >* m_ppPoints;
};




//============================================================================
//    ReadPLYFileHelper inline methods
//============================================================================


template< typename t_Type >
ReadPLYFileHelper< t_Type >::ReadPLYFileHelper() {

	ply::ply_parser::scalar_property_definition_callbacks_type callbacks;
	ply::at< ply::float32 >( callbacks ) = std::tr1::bind( &ReadPLYFileHelper::DefineCallbacks, this, std::tr1::placeholders::_1, std::tr1::placeholders::_2 );
	m_parser.scalar_property_definition_callbacks( callbacks );
}


template< typename t_Type >
void ReadPLYFileHelper< t_Type >::Read(
	unsigned int& dimension,
	std::vector< boost::shared_array< t_Type > >& pPoints,
	boost::filesystem::path const& path
)
{
	m_index = 0;

	m_dimension = 0;
	m_ppPoints = &pPoints;

	std::string const plyFilename = path.string();
	m_parser.parse( plyFilename );

	dimension = m_dimension;
	m_ppPoints = NULL;
}




//============================================================================
//    ReadPLYFileHelper methods
//============================================================================


template< typename t_Type >
std::tr1::function< void ( ply::float32 ) > ReadPLYFileHelper< t_Type >::DefineCallbacks( std::string const& elementName, std::string const& propertyName ) {

	if ( elementName == "vertex" ) {

		if ( propertyName == "x" ) {

			m_dimension = std::max( m_dimension, 1u );
			return std::tr1::bind( &ReadPLYFileHelper::VertexCallback, this, 0, std::tr1::placeholders::_1 );
		}
		else if ( propertyName == "y" ) {

			m_dimension = std::max( m_dimension, 2u );
			return std::tr1::bind( &ReadPLYFileHelper::VertexCallback, this, 1, std::tr1::placeholders::_1 );
		}
		else if ( propertyName == "z" ) {

			m_dimension = std::max( m_dimension, 3u );
			return std::tr1::bind( &ReadPLYFileHelper::VertexCallback, this, 2, std::tr1::placeholders::_1 );
		}
		else if ( propertyName == "w" ) {

			m_dimension = std::max( m_dimension, 4u );
			return std::tr1::bind( &ReadPLYFileHelper::VertexCallback, this, 3, std::tr1::placeholders::_1 );
		}
	}

	return 0;
}


template< typename t_Type >
void ReadPLYFileHelper< t_Type >::VertexCallback( unsigned int index, ply::float32 const value ) {

	BOOST_ASSERT( m_dimension > 0 );
	if ( index != m_index )
		throw std::runtime_error( "Point coordinate found at unexpected index" );

	if ( m_index == 0 )
		m_pPoint = boost::shared_array< t_Type >( new t_Type[ m_dimension ] );

	BOOST_ASSERT( m_pPoint.get() != NULL );
	m_pPoint[ index ] = value;
	++m_index;

	if ( m_index == m_dimension ) {

		( *m_ppPoints ).push_back( m_pPoint );
		m_pPoint = boost::shared_array< t_Type >();
		m_index = 0;
	}
}




}    // anonymous namespace




//============================================================================
//    ReadPLYFile function
//============================================================================


template< typename t_Type >
void ReadPLYFile(
	unsigned int& dimension,
	std::vector< boost::shared_array< t_Type > >& pPoints,
	boost::filesystem::path const& path
)
{
	ReadPLYFileHelper< t_Type > helper;
	helper.Read( dimension, pPoints, path );
}




//============================================================================
//    ReadNodeFile function
//============================================================================


template< typename t_Type >
void ReadNodeFile(
	unsigned int& dimension,
	std::vector< boost::shared_array< t_Type > >& pPoints,
	boost::filesystem::path const& path
)
{
	unsigned int const bufferSize = ( 1u << 16 );
	boost::shared_array< char > buffer( new char[ bufferSize ] );
	boost::regex const spaceRegex( "[[:space:]]+" );
	boost::regex const commentRegex( "#.*$" );
	boost::regex const leadingSpaceRegex(  "^[[:space:]]*" );
	boost::regex const trailingSpaceRegex( "[[:space:]]*$" );

	std::string const filename = path.string();
	FILE* file = fopen( filename.c_str(), "rb" );
	if ( file == NULL )
		throw std::runtime_error( "Unable to open node file" );

	unsigned int points = 0;
	for ( bool done = false; ! done; ) {

		if ( fgets( buffer.get(), bufferSize, file ) == NULL )
			throw std::runtime_error( "Header not found in node file" );

		std::string lineString( buffer.get() );
		lineString = boost::regex_replace( lineString, commentRegex,       "" );
		lineString = boost::regex_replace( lineString, leadingSpaceRegex,  "" );
		lineString = boost::regex_replace( lineString, trailingSpaceRegex, "" );
		if ( lineString != "" ) {

			boost::sregex_token_iterator ii(
				lineString.begin(),
				lineString.end(),
				spaceRegex,
				-1
			);
			boost::sregex_token_iterator iiEnd;

			{	if ( ii == iiEnd )
					throw std::runtime_error( "Number of points not found in node file" );
				std::string const valueString = ii->str();
				int const signedPoints = atoi( valueString.c_str() );
				if ( signedPoints <= 0 )
					throw std::runtime_error( "Number of points is non-positive" );
				points = signedPoints;
			}

			++ii;

			{	if ( ii == iiEnd )
					throw std::runtime_error( "Dimension not found in node file" );
				std::string const valueString = ii->str();
				int const signedDimension = atoi( valueString.c_str() );
				if ( signedDimension <= 0 )
					throw std::runtime_error( "Dimension is non-positive" );
				dimension = signedDimension;
			}

			done = true;
		}
	}

	while ( ( pPoints.size() < points ) && ( ! feof( file ) ) ) {

		if ( fgets( buffer.get(), bufferSize, file ) == NULL )
			break;

		std::string lineString( buffer.get() );
		lineString = boost::regex_replace( lineString, commentRegex,       "" );
		lineString = boost::regex_replace( lineString, leadingSpaceRegex,  "" );
		lineString = boost::regex_replace( lineString, trailingSpaceRegex, "" );
		if ( lineString != "" ) {

			boost::sregex_token_iterator ii(
				lineString.begin(),
				lineString.end(),
				spaceRegex,
				-1
			);
			boost::sregex_token_iterator iiEnd;
			++ii;    // skip label

			boost::shared_array< t_Type > pPoint( new t_Type[ dimension ] );

			for ( unsigned int jj = 0; jj < dimension; ++jj, ++ii ) {

				if ( ii == iiEnd )
					throw std::runtime_error( "Point has wrong dimension" );

				std::string const valueString = ii->str();
				pPoint[ jj ] = atof( valueString.c_str() );
			}
			//if ( ii != iiEnd )
			//	throw std::runtime_error( "Point has wrong dimension" );

			pPoints.push_back( pPoint );
		}
	}

	if ( pPoints.size() < points )
		throw std::runtime_error( "Too few points found in node file" );

	fclose( file );
}




//============================================================================
//    ReadFile function
//============================================================================


template< typename t_Type >
void ReadFile(
	unsigned int& dimension,
	std::vector< boost::shared_array< t_Type > >& pPoints,
	boost::filesystem::path const& path
)
{
	std::string const extension = boost::filesystem::extension( path );
	if ( strcasecmp( extension.c_str(), ".ply" ) == 0 )
		ReadPLYFile( dimension, pPoints, path );
	else if ( ( strcasecmp( extension.c_str(), ".node" ) == 0 ) || ( strcasecmp( extension.c_str(), ".poly" ) == 0 ) )
		ReadNodeFile( dimension, pPoints, path );
	else
		throw std::runtime_error( "File has unknown extension" );
}




#endif    /* __cplusplus */

#endif    /* __DATA_HH__ */
