/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file pool.hh
	\brief Pool and PoolList classes
*/




#ifndef __POOL_HH__
#define __POOL_HH__

#ifdef __cplusplus




#include <map>
#include <vector>
#include <new>




//============================================================================
//    Pool class
//============================================================================


struct Pool {

	inline Pool( unsigned int const size );
	inline Pool( unsigned int const size, unsigned int const blockSize );
	inline ~Pool();


	inline void* const Allocate();
	inline void Free( void* const pData );


private:

	inline void AllocateBlock( unsigned int const blockSize );


	unsigned int m_size;
	unsigned int m_blockSize;
	void* m_pCurrent;
	std::vector< void* > m_pBlocks;
};




//============================================================================
//    Pool inline methods
//============================================================================


Pool::Pool( unsigned int const size ) :
	m_size( ( size + ( sizeof( void* ) - 1 ) ) / sizeof( void* ) ),
	m_blockSize( std::max( 16u, ( 4095 + m_size ) / 4096 ) ),
	m_pCurrent( NULL )
{
}


Pool::Pool( unsigned int const size, unsigned int const blockSize ) :
	m_size( ( size + ( sizeof( void* ) - 1 ) ) / sizeof( void* ) ),
	m_blockSize( blockSize ),
	m_pCurrent( NULL )
{
}


Pool::~Pool() {

	std::vector< void* >::const_iterator ii    = m_pBlocks.begin();
	std::vector< void* >::const_iterator iiEnd = m_pBlocks.end();
	for ( ; ii != iiEnd; ++ii )
		operator delete( *ii );
}


void* const Pool::Allocate() {

	if ( m_pCurrent == NULL ) {

		AllocateBlock( m_blockSize );
		m_blockSize += m_blockSize;
	}
	BOOST_ASSERT( m_pCurrent != NULL );

	void* const pResult = m_pCurrent;
	m_pCurrent = *reinterpret_cast< void** >( m_pCurrent );

	return pResult;
}


void Pool::Free( void* const pData ) {

	*reinterpret_cast< void** >( pData ) = m_pCurrent;
	m_pCurrent = pData;
}


void Pool::AllocateBlock( unsigned int const blockSize ) {

	void* const pBlock = operator new( blockSize * m_size * sizeof( void* ) );
	{	void** ii    = reinterpret_cast< void** >( pBlock );
		void** iiEnd = reinterpret_cast< void** >( pBlock ) + ( blockSize * m_size );
		for ( ; ii < iiEnd; ii += m_size ) {

			*ii = m_pCurrent;
			m_pCurrent = reinterpret_cast< void* >( ii );
		}
		BOOST_ASSERT( ii == iiEnd );
	}

	m_pBlocks.push_back( pBlock );
}




//============================================================================
//    PoolList class
//============================================================================


struct PoolList {

	inline PoolList();
	inline ~PoolList();


	inline Pool* const GetPool( unsigned int const size );


private:

	std::map< unsigned int, Pool* > m_poolMap;
};




//============================================================================
//    PoolList inline methods
//============================================================================


PoolList::PoolList() {
}


PoolList::~PoolList() {

	std::map< unsigned int, Pool* >::iterator ii    = m_poolMap.begin();
	std::map< unsigned int, Pool* >::iterator iiEnd = m_poolMap.end();
	for ( ; ii != iiEnd; ++ii )
		delete ii->second;
}


Pool* const PoolList::GetPool( unsigned int const size ) {

	std::map< unsigned int, Pool* >::iterator ii = m_poolMap.find( size );
	if ( ii == m_poolMap.end() )
		ii = m_poolMap.insert( std::pair< unsigned int, Pool* >( size, new Pool( size ) ) ).first;
	return ii->second;
}




#endif    /* __cplusplus */

#endif    /* __POOL_HH__ */
