/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file quadtree.hh
	\brief Quadtree class
*/




#ifndef __QUADTREE_HH__
#define __QUADTREE_HH__

#ifdef __cplusplus




#include "clipped_voronoi.hh"
#include "vector.hh"
#include "accuracy.hh"
#include "refcount.hh"
#include "pool.hh"
#include "maybe_threaded.hh"

#include <gmpxx.h>

#include <boost/thread.hpp>
#include <boost/function.hpp>
#include <boost/array.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/assert.hpp>
#include <boost/static_assert.hpp>

#include <set>
#include <list>
#include <queue>
#include <algorithm>
#include <limits>
#include <cmath>




namespace {




//============================================================================
//    TrueFunction helper function
//============================================================================


template< typename t_Type >
inline bool const TrueFunction( t_Type const& ) {

	return true;
}




//============================================================================
//    NOPFunction helper function
//============================================================================


template< typename t_Type >
inline void const NOPFunction( t_Type const& ) {
}




}    // anonymous namespace




//============================================================================
//    Quadtree class
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads = 1, typename t_Type = double, typename t_ClientDataType = void >
struct Quadtree {

	BOOST_STATIC_ASSERT( t_Dimension > 0 );
	BOOST_STATIC_ASSERT( t_Threads > 0 );


	enum { DIMENSION = t_Dimension };
	typedef t_Type Type;


private:

	struct Point;
	struct Node;


public:

	// opaque handle to an inserted point (never invalidated)
	typedef Point const* Handle;


	struct NodeList {

		inline NodeList();
		inline ~NodeList();


		inline bool const IsEmpty();

		inline void Clear();


		/*
			Called whenever the corresponding operation occurs in a node in
			this NodeList. It is always safe to call Clear() within a
			callback
		*/
		virtual void SplitNotification();
		virtual void MergeNotification();
		virtual void InsertNotification();
		virtual void InsertSteinerNotification( unsigned int const thread );
		virtual void EraseNotification();
		virtual void EraseSteinerNotification( unsigned int const thread );


	private:

		typename MaybeThreaded< ( t_Threads > 1 ) >::Mutex m_mutex;
		std::set< Node* > m_pNodes;


		friend struct Quadtree;
		friend struct Node;


		// private and unimplemented
		inline NodeList( NodeList const& );
		inline NodeList const& operator=( NodeList const& );
	};


	inline Quadtree( unsigned int const nodeSize, double const delta = 1.0 );
	inline ~Quadtree();


	// get the point coordinates out of a handle
	static inline boost::array< t_Type, t_Dimension > const& GetPoint( Handle const& handle );
	static inline ClientData< t_ClientDataType >& GetClientData( Handle const& handle );


	inline unsigned int const GetPoints() const;
	inline unsigned int const GetSteinerPoints() const;


	inline bool const EnumeratePoints(        boost::function1< bool const, Handle const& > const& enumerator ) const;    // enumerator is Handle -> bool, false iff enumeration should stop
	inline bool const EnumerateSteinerPoints( boost::function1< bool const, Handle const& > const& enumerator ) const;    // enumerator is Handle -> bool, false iff enumeration should stop


	// returns NULL if point is a duplicate of a point already in the quadtree
	template< typename t_Callback >
	inline Handle const Insert(
		boost::array< t_Type, t_Dimension > const& point,
		t_Callback& callback
	);

	// returns NULL if point is a duplicate of a point already in the quadtree
	inline Handle const Insert(
		boost::array< t_Type, t_Dimension > const& point
	);


	// does not check for duplicates
	template< typename t_Callback >
	inline Handle const InsertSteiner(
		boost::array< t_Type, t_Dimension > const& point,
		t_Callback& callback,
		unsigned int const thread = 0
	);

	// does not check for duplicates
	inline Handle const InsertSteiner(
		boost::array< t_Type, t_Dimension > const& point,
		unsigned int const thread = 0
	);


	// memory for the erased point is not freed until CleanupErased() is called
	inline void Erase(
		Handle const& handle
	);

	// memory for the erased point is not freed until CleanupErased() is called
	inline void EraseSteiner(
		Handle const& handle,
		unsigned int const thread = 0
	);

	// frees memory associated with erased points
	inline void CleanupErased();


	// is exact. does *not* include Steiner points
	inline t_Type const LowerBoundNearestNeighborDistanceSquared(
		NodeList* const pNodeList,    // searched node will be added to this list. may be NULL
		Handle const& handle
	) const;

	// is exact. also includes Steiner points
	template< typename t_Functor >
	Handle const FindNearestNeighbor(
		NodeList* const pNodeList,    // searched nodes will be added to this list. may be NULL
		Handle const& handle,
		unsigned int const thread,
		t_Functor& functor    // Handle -> bool, true iff point should be included
	) const;

	// is exact. also includes Steiner points
	inline Handle const FindNearestNeighbor(
		NodeList* const pNodeList,    // searched nodes will be added to this list. may be NULL
		Handle const& handle,
		unsigned int const thread = 0
	) const;

	// is exact. also includes Steiner points
	template< typename t_Functor >
	boost::shared_ptr< ClippedVoronoi< t_Dimension, t_Type, Handle > > const FindClippedVoronoi(
		PoolList& poolList,
		NodeList* const pNodeList,    // searched nodes will be added to this list. may be NULL
		Handle const& handle,
		Accuracy::Exact const& nearestNeighborDistanceSquared,
		double const rho,
		double const beta,
		unsigned int const thread,
		t_Functor& functor    // Handle -> bool, true iff point should be included
	) const;

	// is exact. also includes Steiner points
	inline boost::shared_ptr< ClippedVoronoi< t_Dimension, t_Type, Handle > > const FindClippedVoronoi(
		PoolList& poolList,
		NodeList* const pNodeList,    // searched nodes will be added to this list. may be NULL
		Handle const& handle,
		Accuracy::Exact const& nearestNeighborDistanceSquared,
		double const rho,
		double const beta,
		unsigned int const thread = 0
	) const;


private:

	struct Point : public virtual Refcounted< ( t_Threads > 1 ) >, public boost::array< t_Type, t_Dimension >, public ClientData< t_ClientDataType > {

		inline Point( boost::array< t_Type, t_Dimension > const& data, Node* const pOwner );


		inline Node* const GetOwner() const;

		inline void SetOwner( Node* const pOwner );


	private:

		Node* m_pOwner;


		// private and unimplemented
		inline Point( Point const& );
		inline Point const& operator=( Point const& );
	};


	struct Node {

		inline Node();
		inline Node( Quadtree* const pQuadtree );

		inline ~Node();


		inline bool const IsLeaf() const;

		inline Quadtree* const GetQuadtree() const;
		inline Node* const GetParent() const;

		inline unsigned int const GetPoints() const;
		inline unsigned int const GetSteinerPoints() const;

		inline boost::array< t_Type, t_Dimension > const& GetLower() const;
		inline boost::array< t_Type, t_Dimension > const& GetUpper() const;


		bool const EnumeratePoints(        boost::function1< bool const, Handle const& > const& enumerator ) const;    // enumerator is Handle -> bool, false iff enumeration should stop
		bool const EnumerateSteinerPoints( boost::function1< bool const, Handle const& > const& enumerator ) const;    // enumerator is Handle -> bool, false iff enumeration should stop


		void ClearTag( unsigned int const thread ) const;


		// inserts a point, maintaining the balance property
		template< typename t_Callback >
		Handle const Insert(
			boost::array< t_Type, t_Dimension > const& point,
			t_Callback& callback
		);

		/*
			inserts a Steiner point (to which the balance and crowding
			properties do not apply)
		*/
		template< typename t_Callback >
		Handle const InsertSteiner(
			boost::array< t_Type, t_Dimension > const& point,
			t_Callback& callback,
			unsigned int const thread
		);

		static void Erase(
			Handle const& handle
		);

		static void EraseSteiner(
			Handle const& handle,
			unsigned int const thread
		);


		inline t_Type const LowerBoundNearestNeighborDistanceSquared() const;


	private:

		inline Node* const GetNeighbor( unsigned int const index ) const;

		/*
			maintains the balance property when splitting. returns the leaf
			containing the point, if any, NULL otherwise.
		*/
		void Split();

		/*
			returns true if all points in this node (and its children, etc)
			are further away from the point than the given threshold (uses
			approximate math, so there is actually a bit of wiggle room here.
			the truth is that this function will only return true if it can
			guarantee, using approximate math, that all points satisfy the
			above property)
		*/
		bool const CheckMinimumDistanceSquared(
			boost::array< t_Type, t_Dimension > const& point,
			double const distanceSquaredThreshold
		) const;

		inline bool const IsBoundary( unsigned int const neighborIndex ) const;
		inline bool const CanMerge() const;

		static inline bool const ComparePointsLexicographically( RefcountedPointer< Point > const& pFirst, RefcountedPointer< Point > const& pSecond );


		static void InitializeStatic();


		/*
			Consider the following quadtree (two layers are displayed)
			
			+-------+-------+-------+    +---+---+---+---+---+---+
			|       |       |       |    |   |   |   |   |   |   |
			|       |       |       |    |   |   |   |   |   |   |
			|       |       |       |    +---+---+---+---+---+---+
			|       |       |       |    |   |   |   |   |   |   |
			|       |       |       |    |   |   |   |   |   |   |
			+-------+-------+-------+    +---+---+---+---+---+---+
			|       |       |       |    |   |   |   |   |   |   |
			|       | 0   1 | 2     |    |   |   | 0 | 1 | 2 |   |
			|       |       |       |    +---+---+---+---+---+---+
			|       |       |       |    |   |   |   | * |   |   |
			|       | 3   4 | 5     |    |   |   | 3 | 4 | 5 |   |
			+-------+-------+-------+    +---+---+---+---+---+---+
			|       |       |       |    |   |   |   |   |   |   |
			|       | 6   7 | 8     |    |   |   | 6 | 7 | 8 |   |
			|       |       |       |    +---+---+---+---+---+---+
			|       |       |       |    |   |   |   |   |   |   |
			|       |       |       |    |   |   |   |   |   |   |
			+-------+-------+-------+    +---+---+---+---+---+---+

			The right-hand diagram illustrates the indices of the neighbors
			of the node marked with "*", while the left-hand diagram indicates
			the parent-neighbors. Essentially, the parent-neighbors are the
			nodes one level higher up in the tree which contain the neighbor
			of the same index. The reason for the existence of the
			parent-neighbor pointers is that (except at the boundary of the
			quadtree) parent-neighbors are always guaranteed to exist, while
			child-neighbors are not. Note that, for a 2d quadtree, the 4th
			parent-neighbor is also the parent
		*/


		Quadtree* m_pQuadtree;
		unsigned int m_childIndex;    // index within parent's children

		unsigned int m_points;
		std::vector< RefcountedPointer< Point > > m_pPoints;
		std::vector< RefcountedPointer< Point > > m_pSteinerPoints;

		boost::array< t_Type, t_Dimension > m_lower;
		boost::array< t_Type, t_Dimension > m_upper;

		boost::array< Node, ( 1ull << t_Dimension ) >* m_pChildren;
		Node* m_pParentNeighbors[ Power< 3, t_Dimension >::RESULT ];

		double m_deltaSquared;    // lower bound on the nearest-neighbor distance squared
		unsigned int mutable m_tag[ t_Threads ];

		std::set< NodeList* > m_pNodeLists;

		typename MaybeThreaded< ( t_Threads > 1 ) >::Mutex m_mutex;


		static unsigned int s_neighborIndices[ 1ull << t_Dimension ][ Power< 3, t_Dimension >::RESULT ];
		static unsigned int s_childIndices[    1ull << t_Dimension ][ Power< 3, t_Dimension >::RESULT ];

		static boost::once_flag s_once;


		friend struct Quadtree;
		friend struct NodeList;


		// private and unimplemented
		inline Node( Node const& );
		inline Node const& operator=( Node const& );
	};


	template< typename t_ValueType >
	struct QueueElement {

		inline QueueElement(
			Accuracy::Approximate const& distance,
			t_ValueType const& pNode
		);


		inline Accuracy::Approximate const& GetDistance() const;
		inline t_ValueType const& GetValue() const;


		inline bool const operator==( QueueElement const& other ) const;
		inline bool const operator!=( QueueElement const& other ) const;

		inline bool const operator<( QueueElement const& other ) const;
		inline bool const operator>=( QueueElement const& other ) const;

		inline bool const operator>( QueueElement const& other ) const;
		inline bool const operator<=( QueueElement const& other ) const;


	private:

		Accuracy::Approximate m_distance;
		t_ValueType m_value;
	};


	static inline Accuracy::Approximate const ApproximateEuclideanBoxDistanceSquared(
		boost::array< t_Type, t_Dimension > const& lower,
		boost::array< t_Type, t_Dimension > const& upper,
		boost::array< t_Type, t_Dimension > const& point
	);


	Node* m_pRoot;

	unsigned int m_nodeSize;

	typename MaybeThreaded< ( t_Threads > 1 ) >::Mutex m_mutex;

	std::vector< RefcountedPointer< Point > > m_pErasedPoints;

#ifndef NDEBUG
	unsigned int mutable m_callbacks[ t_Threads ];
#endif    // NDEBUG

	unsigned int mutable m_tag[ t_Threads ];


	friend struct Node;


	// private and unimplemented
	inline Quadtree( Quadtree const& );
	inline Quadtree const& operator=( Quadtree const& );
};




//============================================================================
//    Quadtree::Point inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Point::Point(
	boost::array< t_Type, t_Dimension > const& data,
	Node* const pOwner
) :
	boost::array< t_Type, t_Dimension >( data ),
	m_pOwner( pOwner )
{
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node* const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Point::GetOwner() const {

	BOOST_ASSERT( m_pOwner != NULL );
	return m_pOwner;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Point::SetOwner( Node* const pOwner ) {

	m_pOwner = pOwner;
	BOOST_ASSERT( m_pOwner != NULL );
}




//============================================================================
//    Quadtree::Node static data
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
unsigned int Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::s_neighborIndices[ 1ull << t_Dimension ][ Power< 3, t_Dimension >::RESULT ];


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
unsigned int Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::s_childIndices[ 1ull << t_Dimension ][ Power< 3, t_Dimension >::RESULT ];


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
boost::once_flag Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::s_once = BOOST_ONCE_INIT;




//============================================================================
//    Quadtree::Node inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::Node() :
	m_childIndex( 1ull << t_Dimension ),
	m_points( 0 ),
	m_pChildren( NULL )
{
	std::fill( m_tag, m_tag + t_Threads, 0 );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::Node( Quadtree* const pQuadtree ) :
	m_pQuadtree( pQuadtree ),
	m_childIndex( 1ull << t_Dimension ),
	m_points( 0 ),
	m_pChildren( NULL )
{
	boost::call_once( InitializeStatic, s_once );

	std::fill( m_pParentNeighbors, m_pParentNeighbors + ARRAYLENGTH( m_pParentNeighbors ), static_cast< Node* >( NULL ) );
	std::fill( m_tag, m_tag + t_Threads, 0 );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::~Node() {

	if ( m_pChildren != NULL )
		delete m_pChildren;

	{	typename std::set< NodeList* >::const_iterator ii    = m_pNodeLists.begin();
		typename std::set< NodeList* >::const_iterator iiEnd = m_pNodeLists.end();
		for ( ; ii != iiEnd; ++ii )
			( *ii )->m_pNodes.erase( this );
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::IsLeaf() const {

	return( m_pChildren == NULL );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >* const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::GetQuadtree() const {

	return m_pQuadtree;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node* const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::GetParent() const {

	return m_pParentNeighbors[ ( Power< 3, t_Dimension >::RESULT ) / 2 ];
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
unsigned int const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::GetPoints() const {

	return m_points;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
unsigned int const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::GetSteinerPoints() const {

	unsigned int result = 0;

	if ( IsLeaf() )
		result = m_pSteinerPoints.size();
	else {

		BOOST_ASSERT( m_pSteinerPoints.begin() == m_pSteinerPoints.end() );
		for ( unsigned int ii = 0; ii < ( 1ull << t_Dimension ); ++ii ) {

			BOOST_ASSERT( m_pChildren != NULL );
			result += ( *m_pChildren )[ ii ].GetSteinerPoints();
		}
	}

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
boost::array< t_Type, t_Dimension > const& Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::GetLower() const {

	return m_lower;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
boost::array< t_Type, t_Dimension > const& Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::GetUpper() const {

	return m_upper;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
t_Type const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::LowerBoundNearestNeighborDistanceSquared() const {

	return m_deltaSquared;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node* const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::GetNeighbor( unsigned int const index ) const {

	Node* pNeighbor = const_cast< Node* >( this );
	if ( index != ( Power< 3, t_Dimension >::RESULT ) / 2 ) {

		BOOST_ASSERT( index < ( Power< 3, t_Dimension >::RESULT ) );
		unsigned int const childIndex = s_childIndices[ m_childIndex ][ index ];

		Node* const pParentNeighbor = m_pParentNeighbors[ index ];
		boost::array< Node, ( 1ull << t_Dimension ) >* const pParentNeighborChildren = ( ( pParentNeighbor == NULL ) ? NULL : pParentNeighbor->m_pChildren );
		pNeighbor = ( ( pParentNeighborChildren == NULL ) ? NULL : &( *pParentNeighborChildren )[ childIndex ] );
	}
	return pNeighbor;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::IsBoundary( unsigned int const neighborIndex ) const {

	BOOST_ASSERT( neighborIndex < ( Power< 3, t_Dimension >::RESULT ) );

	bool result = false;
	for ( unsigned int ii = 0, jj = 1; ( ! result ) && ( ii < t_Dimension ); ++ii, jj *= 3 ) {

		switch ( ( neighborIndex / jj ) % 3 ) {
			case 0: result = ( m_lower[ ii ] <= 0 ); break;
			case 2: result = ( m_upper[ ii ] >= 1 ); break;
			default: break;
		}
	}
	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::CanMerge() const {

	bool result = ( m_points <= m_pQuadtree->m_nodeSize );

	// check for balance
	for ( unsigned int ii = 0; result && ( ii < ( Power< 3, t_Dimension >::RESULT ) ); ++ii ) {

		Node const* const pNeighbor = GetNeighbor( ii );
		if ( ( pNeighbor != NULL ) && ( ! pNeighbor->IsLeaf() ) ) {

			// check all of the neighbor's children which adjoin this square

			unsigned int mask = 0;
			unsigned int bits = 0;
			for ( unsigned int jj = 1, kk = 1; jj < ( 1ull << t_Dimension ); jj *= 2, kk *= 3 ) {

				BOOST_ASSERT( kk < ( Power< 3, t_Dimension >::RESULT ) );
				switch ( ( ii / kk ) % 3 ) {

					case 0: mask |= jj; bits |= jj; break;
					case 2: mask |= jj; break;
					default: break;
				}
			}

			for ( unsigned int jj = 0; result && ( jj < ( 1ull << t_Dimension ) ); ++jj ) {

				if ( ( jj & mask ) == bits ) {

					BOOST_ASSERT( pNeighbor->m_pChildren != NULL );
					Node const* const pChild = &( *pNeighbor->m_pChildren )[ jj ];
					BOOST_ASSERT( pChild != NULL );
					result = pChild->IsLeaf();
				}
			}
		}
	}

	// check for crowding
	if ( result && ( m_points > 0 ) ) {

		double const distanceSquaredThreshold = m_deltaSquared;

		std::vector< RefcountedPointer< Point > > pPoints;

		for ( unsigned int ii = 0; ii < ( 1ull << t_Dimension ); ++ii ) {

			BOOST_ASSERT( m_pChildren != NULL );
			Node const* const pChild = &( *m_pChildren )[ ii ];
			BOOST_ASSERT( ( pChild != NULL ) && ( pChild->IsLeaf() ) );

			pPoints.insert( pPoints.end(), pChild->m_pPoints.begin(), pChild->m_pPoints.end() );
		}

		// check points in children
		{	typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pPoints.begin();
			typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pPoints.end();
			for ( ; result && ( ii != iiEnd ); ++ii ) {

				typename std::vector< RefcountedPointer< Point > >::const_iterator jj = ii;
				++jj;
				for ( ; result && ( jj != iiEnd ); ++jj )
					if ( Accuracy::Approximate::Compare( ApproximateEuclideanDistanceSquared( **ii, **jj ), distanceSquaredThreshold ) <= 0 )
						result = false;
			}
		}

		// check points in neighbors
		for ( unsigned int ii = 0; result && ( ii < ( Power< 3, t_Dimension >::RESULT ) ); ++ii ) {

			if ( ii != ( Power< 3, t_Dimension >::RESULT ) / 2 ) {

				Node* const pNeighbor = GetNeighbor( ii );

				if ( ( pNeighbor != NULL ) && ( pNeighbor->m_points > 0 ) ) {

					typename std::vector< RefcountedPointer< Point > >::const_iterator jj    = pPoints.begin();
					typename std::vector< RefcountedPointer< Point > >::const_iterator jjEnd = pPoints.end();
					for ( ; result && ( jj != jjEnd ); ++jj )
						result = pNeighbor->CheckMinimumDistanceSquared( **jj, distanceSquaredThreshold );
				}
			}
		}
	}

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::ComparePointsLexicographically( RefcountedPointer< Point > const& pFirst, RefcountedPointer< Point > const& pSecond ) {

	BOOST_ASSERT( pFirst.Get()  != NULL );
	BOOST_ASSERT( pSecond.Get() != NULL );

	bool result = false;
	for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

		if ( ( *pFirst )[ ii ] < ( *pSecond )[ ii ] ) {

			result = true;
			break;
		}
		else if ( ( *pFirst )[ ii ] > ( *pSecond )[ ii ] ) {

			result = false;
			break;
		}
	}

	return result;
}




//============================================================================
//    Quadtree::Node methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::EnumeratePoints( boost::function1< bool const, Handle const& > const& enumerator ) const {

	bool result = true;

	if ( IsLeaf() ) {

		typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = m_pPoints.begin();
		typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = m_pPoints.end();
		for ( ; result && ( ii != iiEnd ); ++ii )
			result = enumerator( ( *ii ).Get() );
	}
	else {

		for ( unsigned int ii = 0; result && ( ii < ( 1ull << t_Dimension ) ); ++ii ) {

			BOOST_ASSERT( m_pChildren != NULL );
			result = ( *m_pChildren )[ ii ].EnumeratePoints( enumerator );
		}
	}

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::EnumerateSteinerPoints( boost::function1< bool const, Handle const& > const& enumerator ) const {

	bool result = true;

	if ( IsLeaf() ) {

		typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = m_pSteinerPoints.begin();
		typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = m_pSteinerPoints.end();
		for ( ; result && ( ii != iiEnd ); ++ii )
			result = enumerator( ( *ii ).Get() );
	}
	else {

		for ( unsigned int ii = 0; result && ( ii < ( 1ull << t_Dimension ) ); ++ii ) {

			BOOST_ASSERT( m_pChildren != NULL );
			result = ( *m_pChildren )[ ii ].EnumerateSteinerPoints( enumerator );
		}
	}

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::ClearTag( unsigned int const thread ) const {

	m_tag[ thread ] = 0;
	if ( ! IsLeaf() ) {

		for ( unsigned int ii = 0; ii < ( 1ull << t_Dimension ); ++ii ) {

			BOOST_ASSERT( m_pChildren != NULL );
			( *m_pChildren )[ ii ].ClearTag( thread );
		}
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_Callback >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Handle const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::Insert(
	boost::array< t_Type, t_Dimension > const& point,
	t_Callback& callback
)
{
	// must start at the root
	BOOST_ASSERT( GetParent() == NULL );

	// find the insertion point
	Node* pNode = this;
	while ( ! pNode->IsLeaf() ) {

		unsigned int index = 0;
		for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

			t_Type const lower = pNode->m_lower[ ii ];
			t_Type const upper = pNode->m_upper[ ii ];
			t_Type const middle = ( lower + upper ) / 2;

			BOOST_ASSERT( point[ ii ] >= lower );
			BOOST_ASSERT( point[ ii ] <  upper );
			if ( point[ ii ] >= middle )
				index |= ( 1ull << ii );
		}

		BOOST_ASSERT( pNode->m_pChildren != NULL );
		pNode = &( *pNode->m_pChildren )[ index ];
	}

	bool duplicate = false;

	// check points in this node for duplicates
	if ( ! duplicate ) {

		typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pNode->m_pPoints.begin();
		typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pNode->m_pPoints.end();
		for ( ; ii != iiEnd; ++ii ) {

			bool equal = true;
			for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

				if ( point[ jj ] != ( **ii )[ jj ] ) {

					equal = false;
					break;
				}
			}

			if ( equal ) {

				duplicate = true;
				break;
			}
		}
	}

	// check Steiner points in this node for duplicates
	if ( ! duplicate ) {

		typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pNode->m_pSteinerPoints.begin();
		typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pNode->m_pSteinerPoints.end();
		for ( ; ii != iiEnd; ++ii ) {

			bool equal = true;
			for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

				if ( point[ jj ] != ( **ii )[ jj ] ) {

					equal = false;
					break;
				}
			}

			if ( equal ) {

				duplicate = true;
				break;
			}
		}
	}

	Point* pResult = NULL;
	if ( ! duplicate ) {

		// split the insertion point until the bound on node size is satisfied
		while ( pNode->m_points + 1 > m_pQuadtree->m_nodeSize ) {

			pNode->Split();

			unsigned int index = 0;
			for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

				t_Type const lower = pNode->m_lower[ jj ];
				t_Type const upper = pNode->m_upper[ jj ];
				t_Type const middle = ( lower + upper ) / 2;

				BOOST_ASSERT( point[ jj ] >= lower );
				BOOST_ASSERT( point[ jj ] <  upper );
				if ( point[ jj ] >= middle )
					index |= ( 1ull << jj );
			}

			BOOST_ASSERT( pNode->m_pChildren != NULL );
			pNode = &( *pNode->m_pChildren )[ index ];
		}

		/*
			split the insertion point and neighbors until the crowding property is
			satisfied
		*/
		for ( bool done = false; ! done; ) {

			BOOST_ASSERT( pNode != NULL );
			BOOST_ASSERT( pNode->IsLeaf() );

			done = true;
			bool split = false;

			double const distanceSquaredThreshold = pNode->m_deltaSquared;

			// check crowding at parent's neighbors
			for ( unsigned int ii = 0; ii < ( Power< 3, t_Dimension >::RESULT ); ++ii ) {

				Node* const pParentNeighbor = pNode->m_pParentNeighbors[ ii ];

				if ( ( pParentNeighbor != NULL ) && pParentNeighbor->IsLeaf() && ( pParentNeighbor->m_points > 0 ) ) {

					double const neighborDistanceSquaredThreshold = pParentNeighbor->m_deltaSquared;
					BOOST_ASSERT( neighborDistanceSquaredThreshold > distanceSquaredThreshold );

					typename std::vector< RefcountedPointer< Point > >::const_iterator jj    = pParentNeighbor->m_pPoints.begin();
					typename std::vector< RefcountedPointer< Point > >::const_iterator jjEnd = pParentNeighbor->m_pPoints.end();
					for ( ; jj != jjEnd; ++jj ) {

						if ( Accuracy::Approximate::Compare( ApproximateEuclideanDistanceSquared( point, **jj ), neighborDistanceSquaredThreshold ) <= 0 ) {

							pParentNeighbor->Split();
							done = false;

							break;
						}
					}
				}
			}

			// check crowding at neighbors (including self)
			for ( unsigned int ii = 0; ii < ( Power< 3, t_Dimension >::RESULT ); ++ii ) {

				Node* const pNeighbor = pNode->GetNeighbor( ii );

				if ( ( pNeighbor != NULL ) && pNeighbor->IsLeaf() && ( pNeighbor->m_points > 0 ) ) {

					BOOST_ASSERT( distanceSquaredThreshold == pNeighbor->m_deltaSquared );

					typename std::vector< RefcountedPointer< Point > >::const_iterator jj    = pNeighbor->m_pPoints.begin();
					typename std::vector< RefcountedPointer< Point > >::const_iterator jjEnd = pNeighbor->m_pPoints.end();
					for ( ; jj != jjEnd; ++jj ) {

						if ( Accuracy::Approximate::Compare( ApproximateEuclideanDistanceSquared( point, **jj ), distanceSquaredThreshold ) <= 0 ) {

							if ( pNeighbor != pNode )
								pNeighbor->Split();
							split = true;
							done = false;

							break;
						}
					}
				}
			}

			// check crowding at neighbor's children
			for ( unsigned int ii = 0; ( ! split ) && ( ii < ( Power< 3, t_Dimension >::RESULT ) ); ++ii ) {

				Node* const pNeighbor = pNode->GetNeighbor( ii );

				if ( ( pNeighbor != NULL ) && ( ! pNeighbor->IsLeaf() ) && ( pNeighbor->m_points > 0 ) ) {

					if ( ! pNeighbor->CheckMinimumDistanceSquared( point, distanceSquaredThreshold ) ) {

						split = true;
						done = false;
					}
				}
			}

			if ( split ) {

				pNode->Split();

				unsigned int index = 0;
				for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

					t_Type const lower = pNode->m_lower[ jj ];
					t_Type const upper = pNode->m_upper[ jj ];
					t_Type const middle = ( lower + upper ) / 2;

					BOOST_ASSERT( point[ jj ] >= lower );
					BOOST_ASSERT( point[ jj ] <  upper );
					if ( point[ jj ] >= middle )
						index |= ( 1ull << jj );
				}

				BOOST_ASSERT( pNode->m_pChildren != NULL );
				pNode = &( *pNode->m_pChildren )[ index ];
			}
		}

		// create and insert the Point structure
		pResult = new Point( point, pNode );
		callback( pResult );
		pNode->m_pPoints.push_back( RefcountedPointer< Point >( pResult ) );
		for ( Node* pNode2 = pNode; pNode2 != NULL; pNode2 = pNode2->GetParent() )
			++pNode2->m_points;

		// call InsertNotification on NodeLists
		{	typename std::set< NodeList* >::const_iterator ii    = pNode->m_pNodeLists.begin();
			typename std::set< NodeList* >::const_iterator iiEnd = pNode->m_pNodeLists.end();
			while ( ii != iiEnd ) {

				typename std::set< NodeList* >::const_iterator const iiCurrent = ii;
				++ii;

				( *iiCurrent )->InsertNotification();
			}
		}
	}

	return pResult;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_Callback >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Handle const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::InsertSteiner(
	boost::array< t_Type, t_Dimension > const& point,
	t_Callback& callback,
	unsigned int const thread
)
{
	BOOST_ASSERT( GetParent() == NULL );

	// find the insertion point
	Node* pNode = this;
	while ( ! pNode->IsLeaf() ) {

		unsigned int index = 0;
		for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

			t_Type const lower = pNode->m_lower[ ii ];
			t_Type const upper = pNode->m_upper[ ii ];
			t_Type const middle = ( lower + upper ) / 2;

			BOOST_ASSERT( point[ ii ] >= lower );
			BOOST_ASSERT( point[ ii ] <  upper );
			if ( point[ ii ] >= middle )
				index |= ( 1ull << ii );
		}

		BOOST_ASSERT( pNode->m_pChildren != NULL );
		pNode = &( *pNode->m_pChildren )[ index ];
	}

#ifndef NDEBUG
	// check points in this node for duplicates
	{	typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pNode->m_pPoints.begin();
		typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pNode->m_pPoints.end();
		for ( ; ii != iiEnd; ++ii ) {

			bool equal = true;
			for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

				if ( point[ jj ] != ( **ii )[ jj ] ) {

					equal = false;
					break;
				}
			}

			BOOST_ASSERT( ! equal );
		}
	}
#endif    // NDEBUG

#ifndef NDEBUG
	// check Steiner points in this node for duplicates
	{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );

		typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pNode->m_pSteinerPoints.begin();
		typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pNode->m_pSteinerPoints.end();
		for ( ; ii != iiEnd; ++ii ) {

			bool equal = true;
			for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

				if ( point[ jj ] != ( **ii )[ jj ] ) {

					equal = false;
					break;
				}
			}

			BOOST_ASSERT( ! equal );
		}
	}
#endif    // NDEBUG

	// create and insert the Point structure
	Point* const pResult = new Point( point, pNode );
	callback( pResult );
	std::vector< NodeList* > pNodeLists;
	{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );
		pNode->m_pSteinerPoints.push_back( RefcountedPointer< Point >( pResult ) );
		pNodeLists.insert( pNodeLists.begin(), pNode->m_pNodeLists.begin(), pNode->m_pNodeLists.end() );
	}

	// call InsertSteinerNotification on NodeLists
	{	typename std::vector< NodeList* >::const_iterator ii    = pNodeLists.begin();
		typename std::vector< NodeList* >::const_iterator iiEnd = pNodeLists.end();
		for ( ; ii != iiEnd; ++ii )
			( *ii )->InsertSteinerNotification( thread );
	}

	return pResult;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::Erase(
	Handle const& handle
)
{
	Node* const pNode = handle->GetOwner();
	BOOST_ASSERT( pNode != NULL );
	BOOST_ASSERT( pNode->IsLeaf() );

	// search for a match among the points, and erase it
	bool erased = false;
	{	typename std::vector< RefcountedPointer< Point > >::iterator ii = pNode->m_pPoints.begin();
		for ( ; ; ++ii ) {

			BOOST_ASSERT( ii != pNode->m_pPoints.end() );
			if ( *ii == handle ) {

				pNode->m_pQuadtree->m_pErasedPoints.push_back( *ii );
				pNode->m_pPoints.erase( ii );
				erased = true;
				break;
			}
		}
	}
	if ( ! erased )
		throw std::runtime_error( "Unable to find point" );

	for ( Node* pNode2 = pNode; pNode2 != NULL; pNode2 = pNode2->GetParent() )
		--pNode2->m_points;

	// call EraseNotification on NodeLists
	{	typename std::set< NodeList* >::const_iterator ii    = pNode->m_pNodeLists.begin();
		typename std::set< NodeList* >::const_iterator iiEnd = pNode->m_pNodeLists.end();
		while ( ii != iiEnd ) {

			typename std::set< NodeList* >::const_iterator const iiCurrent = ii;
			++ii;

			( *iiCurrent )->EraseNotification();
		}
	}

	// try to merge nodes, from smallest to largest
	std::priority_queue< std::pair< int, Node* > > queue;

	/*
		Push all of the non-leaf nodes which could have been crowded
		by this node onto the queue. Looks at the node itself (which
		is of course a leaf), all of its neighbors, its parent, all of
		its parent's neighbors, and so on
	*/
	{	int depth = 0;
		for ( Node* pNode2 = pNode; pNode2 != NULL; pNode2 = pNode2->GetParent() ) {

			for ( unsigned int ii = 0; ii < ( Power< 3, t_Dimension >::RESULT ); ++ii ) {

				Node* const pNeighbor = pNode2->GetNeighbor( ii );
				if ( ( pNeighbor != NULL ) && ( ! pNeighbor->IsLeaf() ) )
					queue.push( std::pair< int, Node* >( depth, pNeighbor ) );
			}

			--depth;
		}
	}

	while ( ! queue.empty() ) {

		std::pair< int, Node* > top = queue.top();
		do { queue.pop(); } while ( ( ! queue.empty() ) && ( queue.top() == top ) );    // we may have duplicates in the queue

		int const depth = top.first;
		Node* const pNode2 = top.second;

		if ( pNode2->CanMerge() ) {

			// merge the node's children
			for ( unsigned int ii = 0; ii < ( 1ull << t_Dimension ); ++ii ) {

				BOOST_ASSERT( pNode2->m_pChildren != NULL );
				Node* const pChild = &( *pNode2->m_pChildren )[ ii ];
				BOOST_ASSERT( pChild->IsLeaf() );

				// reparent points
				while ( ! pChild->m_pPoints.empty() ) {

					RefcountedPointer< Point > const pPoint = pChild->m_pPoints.back();
					pChild->m_pPoints.pop_back();

					pPoint->SetOwner( pNode2 );
					pNode2->m_pPoints.push_back( pPoint );
				}
				while ( ! pChild->m_pSteinerPoints.empty() ) {

					RefcountedPointer< Point > const pPoint = pChild->m_pSteinerPoints.back();
					pChild->m_pSteinerPoints.pop_back();

					pPoint->SetOwner( pNode2 );
					pNode2->m_pSteinerPoints.push_back( pPoint );
				}

				// add child NodeLists to parent NodeList
				pNode2->m_pNodeLists.insert( pChild->m_pNodeLists.begin(), pChild->m_pNodeLists.end() );

				// destroy child NodeLists
				{	typename std::set< NodeList* >::const_iterator jj    = pChild->m_pNodeLists.begin();
					typename std::set< NodeList* >::const_iterator jjEnd = pChild->m_pNodeLists.end();
					for ( ; jj != jjEnd; ++jj )
						( *jj )->m_pNodes.erase( pChild );
				}
				pChild->m_pNodeLists.clear();
			}

			delete pNode2->m_pChildren;
			pNode2->m_pChildren = NULL;
			BOOST_ASSERT( pNode2->IsLeaf() );

			// notify NodeLists of new node
			{	typename std::set< NodeList* >::const_iterator ii    = pNode2->m_pNodeLists.begin();
				typename std::set< NodeList* >::const_iterator iiEnd = pNode2->m_pNodeLists.end();
				for ( ; ii != iiEnd; ++ii )
					( *ii )->m_pNodes.insert( pNode2 );
			}

			// call MergeNotification on NodeLists
			{	typename std::set< NodeList* >::const_iterator ii    = pNode2->m_pNodeLists.begin();
				typename std::set< NodeList* >::const_iterator iiEnd = pNode2->m_pNodeLists.end();
				while ( ii != iiEnd ) {

					typename std::set< NodeList* >::const_iterator const iiCurrent = ii;
					++ii;

					( *iiCurrent )->MergeNotification();
				}
			}

			/*
				Push the non-leaf nodes which could have been smoothed
				by this node onto the queue. Looks at the neighbors of
				the parent
			*/
			Node* const pParent = pNode2->GetParent();
			if ( pParent != NULL ) {

				for ( unsigned int ii = 0; ii < ( Power< 3, t_Dimension >::RESULT ); ++ii ) {

					Node* const pNeighbor = pParent->GetNeighbor( ii );
					if ( ( pNeighbor != NULL ) && ( ! pNeighbor->IsLeaf() ) )
						queue.push( std::pair< int, Node* >( depth - 1, pNeighbor ) );
				}
			}
		}
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::EraseSteiner(
	Handle const& handle,
	unsigned int const thread
)
{
	Node* const pNode = handle->GetOwner();
	BOOST_ASSERT( pNode != NULL );
	BOOST_ASSERT( pNode->IsLeaf() );

	// search for a match among the Steiner points, and erase it
	std::vector< NodeList* > pNodeLists;
	{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );

		typename std::vector< RefcountedPointer< Point > >::iterator ii = pNode->m_pSteinerPoints.begin();
		for ( ; ; ++ii ) {

			BOOST_ASSERT( ii != pNode->m_pSteinerPoints.end() );
			if ( *ii == handle ) {

				{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock quadtreeLock( pNode->m_pQuadtree->m_mutex );
					pNode->m_pQuadtree->m_pErasedPoints.push_back( *ii );
				}

				pNode->m_pSteinerPoints.erase( ii );
				break;
			}
		}

		pNodeLists.insert( pNodeLists.begin(), pNode->m_pNodeLists.begin(), pNode->m_pNodeLists.end() );
	}

	// call EraseSteinerNotification on NodeLists
	{	typename std::vector< NodeList* >::const_iterator ii    = pNodeLists.begin();
		typename std::vector< NodeList* >::const_iterator iiEnd = pNodeLists.end();
		for ( ; ii != iiEnd; ++ii )
			( *ii )->EraseSteinerNotification( thread );
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::Split() {

	BOOST_ASSERT( IsLeaf() );

	std::vector< Node* > pNodes;    // contains nodes to split, from smallest to largest
	pNodes.push_back( this );

	// find all nodes which will need to be split to preserve balance
	for ( unsigned int ii = 0; ii < pNodes.size(); ++ii ) {

		Node* const pNode = pNodes[ ii ];

		if ( pNode->GetParent() != NULL ) {

			// check parent's neighbors in all directions which exit the current square
			for ( unsigned int ii = 1; ii < ( 1ull << t_Dimension ); ++ii ) {

				unsigned int neighborIndex = 0;
				for ( unsigned int jj = 0, kk = 1; jj < t_Dimension; ++jj, kk *= 3 ) {

					if ( ii & ( 1ull << jj ) )
						neighborIndex += ( ( pNode->m_childIndex >> jj ) & 1 ) * 2 * kk;
					else
						neighborIndex += kk;
				}
				BOOST_ASSERT( neighborIndex != ( Power< 3, t_Dimension >::RESULT ) / 2 );    // note that ii starts at 1, not 0

				Node* const pNeighbor = pNode->GetParent()->GetNeighbor( neighborIndex );
				if ( pNeighbor != NULL ) {

					if ( pNeighbor->IsLeaf() )
						pNodes.push_back( pNeighbor );
				}
				else
					BOOST_ASSERT( pNode->GetParent()->IsBoundary( neighborIndex ) );
			}
		}
	}

	// split the nodes in reverse order (from largest to smallest)
	typename std::vector< Node* >::const_reverse_iterator ii    = pNodes.rbegin();
	typename std::vector< Node* >::const_reverse_iterator iiEnd = pNodes.rend();
	for ( ; ii != iiEnd; ++ii ) {

		Node* const pNode = *ii;
		if ( pNode->IsLeaf() ) {    // the node may have already been split (nodes can occur multiple times)

			// call SplitNotification on NodeLists
			{	typename std::set< NodeList* >::const_iterator jj    = pNode->m_pNodeLists.begin();
				typename std::set< NodeList* >::const_iterator jjEnd = pNode->m_pNodeLists.end();
				while ( jj != jjEnd ) {

					typename std::set< NodeList* >::const_iterator const jjCurrent = jj;
					++jj;

					( *jjCurrent )->SplitNotification();
				}
			}

			// create the children
			BOOST_ASSERT( pNode->m_pChildren == NULL );
			pNode->m_pChildren = new boost::array< Node, ( 1ull << t_Dimension ) >;

			// initialize the children
			for ( unsigned int jj = 0; jj < ( 1ull << t_Dimension ); ++jj ) {

				BOOST_ASSERT( pNode->m_pChildren != NULL );
				Node* const pChild = &( *pNode->m_pChildren )[ jj ];

				pChild->m_pQuadtree = pNode->m_pQuadtree;
				pChild->m_childIndex = jj;

				// initialize the bounds
				for ( unsigned int kk = 0; kk < t_Dimension; ++kk ) {

					t_Type const lower = pNode->m_lower[ kk ];
					t_Type const upper = pNode->m_upper[ kk ];
					t_Type const middle = ( lower + upper ) / 2;

					if ( jj & ( 1ull << kk ) ) {

						pChild->m_lower[ kk ] = middle;
						pChild->m_upper[ kk ] = upper;
					}
					else {

						pChild->m_lower[ kk ] = lower;
						pChild->m_upper[ kk ] = middle;
					}
				}
				pChild->m_deltaSquared = 0.25 * pNode->m_deltaSquared;

				// initialize the neighbor and parent-neighbor pointers
				for ( unsigned int kk = 0; kk < Power< 3, t_Dimension >::RESULT; ++kk ) {

					unsigned int const neighborIndex = s_neighborIndices[ jj ][ kk ];

					Node* const pParentNeighbor = pNode->GetNeighbor( neighborIndex );
					BOOST_ASSERT( ( pParentNeighbor != NULL ) || pNode->IsBoundary( neighborIndex ) );

					pChild->m_pParentNeighbors[ kk ] = pParentNeighbor;
				}
			}

			// push the points into the appropriate children
			while ( ! pNode->m_pPoints.empty() ) {

				RefcountedPointer< Point > const pPoint = pNode->m_pPoints.back();
				pNode->m_pPoints.pop_back();

				unsigned int index = 0;
				for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

					t_Type const lower = pNode->m_lower[ jj ];
					t_Type const upper = pNode->m_upper[ jj ];
					t_Type const middle = ( lower + upper ) / 2;

					BOOST_ASSERT( ( *pPoint )[ jj ] >= lower );
					BOOST_ASSERT( ( *pPoint )[ jj ] <  upper );
					if ( ( *pPoint )[ jj ] >= middle )
						index |= ( 1ull << jj );
				}
				++( *pNode->m_pChildren )[ index ].m_points;
				pPoint->SetOwner( &( *pNode->m_pChildren )[ index ] );
				( *pNode->m_pChildren )[ index ].m_pPoints.push_back( pPoint );
			}

			// push the Steiner points into the appropriate children
			while ( ! pNode->m_pSteinerPoints.empty() ) {

				RefcountedPointer< Point > const pPoint = pNode->m_pSteinerPoints.back();
				pNode->m_pSteinerPoints.pop_back();

				unsigned int index = 0;
				for ( unsigned int jj = 0; jj < t_Dimension; ++jj ) {

					t_Type const lower = pNode->m_lower[ jj ];
					t_Type const upper = pNode->m_upper[ jj ];
					t_Type const middle = ( lower + upper ) / 2;

					BOOST_ASSERT( ( *pPoint )[ jj ] >= lower );
					BOOST_ASSERT( ( *pPoint )[ jj ] <  upper );
					if ( ( *pPoint )[ jj ] >= middle )
						index |= ( 1ull << jj );
				}
				pPoint->SetOwner( &( *pNode->m_pChildren )[ index ] );
				( *pNode->m_pChildren )[ index ].m_pSteinerPoints.push_back( pPoint );
			}

			// update NodeLists
			for ( unsigned int jj = 0; jj < ( 1ull << t_Dimension ); ++jj )
				( *pNode->m_pChildren )[ jj ].m_pNodeLists.insert( pNode->m_pNodeLists.begin(), pNode->m_pNodeLists.end() );
			{	typename std::set< NodeList* >::const_iterator jj    = pNode->m_pNodeLists.begin();
				typename std::set< NodeList* >::const_iterator jjEnd = pNode->m_pNodeLists.end();
				for ( ; jj != jjEnd; ++jj ) {

					( *jj )->m_pNodes.erase( pNode );
					for ( unsigned int kk = 0; kk < ( 1ull << t_Dimension ); ++kk )
						( *jj )->m_pNodes.insert( &( *pNode->m_pChildren )[ kk ] );
				}
			}
			pNode->m_pNodeLists.clear();
		}
	}
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::CheckMinimumDistanceSquared(
	boost::array< t_Type, t_Dimension > const& point,
	double const distanceSquaredThreshold
) const
{
	bool result = true;

	if ( m_points > 0 ) {

		if (
			( m_points <= 1 ) ||    // don't bother to use the box bound for a single point
			( Accuracy::Approximate::Compare( ApproximateEuclideanBoxDistanceSquared( m_lower, m_upper, point ), distanceSquaredThreshold ) <= 0 )
		)
		{
			if ( IsLeaf() ) {

				typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = m_pPoints.begin();
				typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = m_pPoints.end();
				for ( ; result && ( ii != iiEnd ); ++ii )
					result = ( Accuracy::Approximate::Compare( ApproximateEuclideanDistanceSquared( point, **ii ), distanceSquaredThreshold ) > 0 );
			}
			else {

				for ( unsigned int childIndex = 0; result && ( childIndex < ( 1ull << t_Dimension ) ); ++childIndex )
					result = ( *m_pChildren )[ childIndex ].CheckMinimumDistanceSquared( point, distanceSquaredThreshold );
			}
		}
	}

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Node::InitializeStatic() {

	for ( unsigned int childIndex = 0; childIndex < ( 1ull << t_Dimension ); ++childIndex ) {

		for ( unsigned int neighborIndex = 0; neighborIndex < ( Power< 3, t_Dimension >::RESULT ); ++neighborIndex ) {

			unsigned int neighborIndex2 = 0;
			unsigned int childIndex2    = 0;
			for ( unsigned int ii = 1, jj = 1; ii < ( 1ull << t_Dimension ); ii *= 2, jj *= 3 ) {

				BOOST_ASSERT( jj < ( Power< 3, t_Dimension >::RESULT ) );
				switch ( ( neighborIndex / jj ) % 3 ) {

					case 0: {

						if ( childIndex & ii ) {

							neighborIndex2 += jj;
							//childIndex2 |= 0;
						}
						else {

							//neighborIndex2 += jj * 0;
							childIndex2 |= ii;
						}

						break;
					}

					case 1: {

						neighborIndex2 += jj;
						childIndex2 |= ( childIndex & ii );

						break;
					}

					case 2: {

						if ( childIndex & ii ) {

							neighborIndex2 += jj * 2;
							// childIndex2 |= 0;
						}
						else {

							neighborIndex2 += jj;
							childIndex2 |= ii;
						}

						break;
					}
				}
			}
			BOOST_ASSERT( neighborIndex2 < ( Power< 3, t_Dimension >::RESULT ) );
			BOOST_ASSERT( childIndex2 < ( 1ull << t_Dimension ) );

			s_neighborIndices[ childIndex ][ neighborIndex ] = neighborIndex2;
			s_childIndices[    childIndex ][ neighborIndex ] = childIndex2;
		}
	}
}




//============================================================================
//    Quadtree::NodeList inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::NodeList::NodeList() {
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::NodeList::~NodeList() {

	Clear();
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::NodeList::IsEmpty() {

	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeListLock( m_mutex );

	return( m_pNodes.empty() );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::NodeList::Clear() {

	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeListLock( m_mutex );

	typename std::set< Node* >::const_iterator ii    = m_pNodes.begin();
	typename std::set< Node* >::const_iterator iiEnd = m_pNodes.end();
	for ( ; ii != iiEnd; ++ii ) {

		typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( ( *ii )->m_mutex );
		( *ii )->m_pNodeLists.erase( this );
	}
	m_pNodes.clear();
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::NodeList::SplitNotification() {
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::NodeList::MergeNotification() {
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::NodeList::InsertNotification() {
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::NodeList::InsertSteinerNotification( unsigned int const thread ) {
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::NodeList::EraseNotification() {
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::NodeList::EraseSteinerNotification( unsigned int const thread ) {
}




//============================================================================
//    Quadtree::QueueElement inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_ValueType >
Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::QueueElement< t_ValueType >::QueueElement(
	Accuracy::Approximate const& distance,
	t_ValueType const& value
) :
	m_distance( distance ),
	m_value( value )
{
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_ValueType >
Accuracy::Approximate const& Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::QueueElement< t_ValueType >::GetDistance() const {

	return m_distance;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_ValueType >
t_ValueType const& Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::QueueElement< t_ValueType >::GetValue() const {

	return m_value;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_ValueType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::QueueElement< t_ValueType >::operator==( QueueElement const& other ) const {

	return( m_distance.GetValue() == other.m_distance.GetValue() );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_ValueType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::QueueElement< t_ValueType >::operator!=( QueueElement const& other ) const {

	return( m_distance.GetValue() != other.m_distance.GetValue() );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_ValueType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::QueueElement< t_ValueType >::operator<( QueueElement const& other ) const {

	return( m_distance.GetValue() > other.m_distance.GetValue() );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_ValueType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::QueueElement< t_ValueType >::operator>=( QueueElement const& other ) const {

	return( m_distance.GetValue() <= other.m_distance.GetValue() );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_ValueType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::QueueElement< t_ValueType >::operator>( QueueElement const& other ) const {

	return( m_distance.GetValue() < other.m_distance.GetValue() );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_ValueType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::QueueElement< t_ValueType >::operator<=( QueueElement const& other ) const {

	return( m_distance.GetValue() >= other.m_distance.GetValue() );
}




//============================================================================
//    Quadtree inline methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Quadtree(
	unsigned int const nodeSize,
	double const delta
) :
	m_pRoot( new Node( this ) ),
	m_nodeSize( nodeSize )
{
	BOOST_ASSERT( m_nodeSize > 0 );

	BOOST_ASSERT( delta >  0 );
	BOOST_ASSERT( delta <= 1 );

	// initialize the bounds
	std::fill( m_pRoot->m_lower.begin(), m_pRoot->m_lower.end(), 0 );
	std::fill( m_pRoot->m_upper.begin(), m_pRoot->m_upper.end(), 1 );

	m_pRoot->m_deltaSquared = Square( delta );

#ifndef NDEBUG
	std::fill( m_callbacks, m_callbacks + t_Threads, 0 );
#endif    // NDEBUG
	std::fill( m_tag, m_tag + t_Threads, 0 );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::~Quadtree() {

	delete m_pRoot;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
boost::array< t_Type, t_Dimension > const& Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::GetPoint( Handle const& handle ) {

	BOOST_ASSERT( handle != NULL );
	return *handle;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
ClientData< t_ClientDataType >& Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::GetClientData( Handle const& handle ) {

	BOOST_ASSERT( handle != NULL );
	return *const_cast< Point* >( handle );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
unsigned int const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::GetPoints() const {

	return m_pRoot->GetPoints();
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
unsigned int const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::GetSteinerPoints() const {

	return m_pRoot->GetSteinerPoints();
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::EnumeratePoints( boost::function1< bool const, Handle const& > const& enumerator ) const {

#ifndef NDEBUG
	++m_callbacks[ 0 ];
#endif    // NDEBUG

	bool const result = m_pRoot->EnumeratePoints( enumerator );

#ifndef NDEBUG
	BOOST_ASSERT( m_callbacks[ 0 ] > 0 );
	--m_callbacks[ 0 ];
#endif    // NDEBUG

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
bool const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::EnumerateSteinerPoints( boost::function1< bool const, Handle const& > const& enumerator ) const {

#ifndef NDEBUG
	++m_callbacks[ 0 ];
#endif    // NDEBUG

	bool const result = m_pRoot->EnumerateSteinerPoints( enumerator );

#ifndef NDEBUG
	BOOST_ASSERT( m_callbacks[ 0 ] > 0 );
	--m_callbacks[ 0 ];
#endif    // NDEBUG

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_Callback >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Handle const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Insert(
	boost::array< t_Type, t_Dimension > const& point,
	t_Callback& callback
)
{
#ifndef NDEBUG
	BOOST_ASSERT( m_callbacks[ 0 ] == 0 );
	++m_callbacks[ 0 ];
#endif    // NDEBUG

	Handle const handle = m_pRoot->Insert( point, callback );

#ifndef NDEBUG
	BOOST_ASSERT( m_callbacks[ 0 ] == 1 );
	--m_callbacks[ 0 ];
#endif    // NDEBUG

	return handle;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Handle const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Insert(
	boost::array< t_Type, t_Dimension > const& point
)
{
	return Insert( point, NOPFunction< Handle > );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_Callback >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Handle const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::InsertSteiner(
	boost::array< t_Type, t_Dimension > const& point,
	t_Callback& callback,
	unsigned int const thread
)
{
	BOOST_ASSERT( thread < t_Threads );

#ifndef NDEBUG
	BOOST_ASSERT( m_callbacks[ thread ] == 0 );
	++m_callbacks[ thread ];
#endif    // NDEBUG

	Handle const handle = m_pRoot->InsertSteiner( point, callback, thread );

#ifndef NDEBUG
	BOOST_ASSERT( m_callbacks[ thread ] == 1 );
	--m_callbacks[ thread ];
#endif    // NDEBUG

	return handle;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Handle const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::InsertSteiner(
	boost::array< t_Type, t_Dimension > const& point,
	unsigned int const thread
)
{
	return InsertSteiner( point, NOPFunction< Handle >, thread );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Erase(
	Handle const& handle
)
{
#ifndef NDEBUG
	BOOST_ASSERT( m_callbacks[ 0 ] == 0 );
	++m_callbacks[ 0 ];
#endif    // NDEBUG

	Node::Erase( handle );

#ifndef NDEBUG
	BOOST_ASSERT( m_callbacks[ 0 ] == 1 );
	--m_callbacks[ 0 ];
#endif    // NDEBUG
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::EraseSteiner(
	Handle const& handle,
	unsigned int const thread
)
{
	BOOST_ASSERT( thread < t_Threads );

#ifndef NDEBUG
	BOOST_ASSERT( m_callbacks[ thread ] == 0 );
	++m_callbacks[ thread ];
#endif    // NDEBUG

	Node::EraseSteiner( handle, thread );

#ifndef NDEBUG
	BOOST_ASSERT( m_callbacks[ thread ] == 1 );
	--m_callbacks[ thread ];
#endif    // NDEBUG
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
void Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::CleanupErased() {

	m_pErasedPoints.clear();
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
t_Type const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::LowerBoundNearestNeighborDistanceSquared(
	NodeList* const pNodeList,
	Handle const& handle
) const
{
	BOOST_ASSERT( handle != NULL );
	Node* const pNode = handle->GetOwner();
	BOOST_ASSERT( pNode != NULL );
	BOOST_ASSERT( pNode->IsLeaf() );

	// pNode is a reader
	if ( pNodeList != NULL ) {

		typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeListLock( pNodeList->m_mutex );
		typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );

		pNodeList->m_pNodes.insert( pNode );
		pNode->m_pNodeLists.insert( pNodeList );
	}

	return pNode->LowerBoundNearestNeighborDistanceSquared();
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Handle const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::FindNearestNeighbor(
	NodeList* const pNodeList,
	Handle const& handle,
	unsigned int const thread
) const
{
	return FindNearestNeighbor( pNodeList, handle, thread, TrueFunction< Handle > );
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
boost::shared_ptr< ClippedVoronoi< t_Dimension, t_Type, typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Handle > > const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::FindClippedVoronoi(
	PoolList& poolList,
	NodeList* const pNodeList,
	Handle const& handle,
	Accuracy::Exact const& nearestNeighborDistanceSquared,
	double const rho,
	double const beta,
	unsigned int const thread
) const
{
	return FindClippedVoronoi(
		poolList,
		pNodeList,
		handle,
		nearestNeighborDistanceSquared,
		rho,
		beta,
		thread,
		TrueFunction< Handle >
	);
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
Accuracy::Approximate const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::ApproximateEuclideanBoxDistanceSquared(
	boost::array< t_Type, t_Dimension > const& lower,
	boost::array< t_Type, t_Dimension > const& upper,
	boost::array< t_Type, t_Dimension > const& point
)
{
	Accuracy::Approximate result( 0 );
	for ( unsigned int ii = 0; ii < t_Dimension; ++ii ) {

		if ( point[ ii ] < lower[ ii ] )
			result += Square( Accuracy::Approximate( lower[ ii ] ) - Accuracy::Approximate( point[ ii ] ) );
		else if ( point[ ii ] > upper[ ii ] )
			result += Square( Accuracy::Approximate( point[ ii ] ) - Accuracy::Approximate( upper[ ii ] ) );
	}

	return result;
}





//============================================================================
//    Quadtree methods
//============================================================================


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_Functor >
typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Handle const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::FindNearestNeighbor(
	NodeList* const pNodeList,
	Handle const& handle,
	unsigned int const thread,
	t_Functor& functor
) const
{
	BOOST_ASSERT( thread < t_Threads );

	Handle result = NULL;

	BOOST_ASSERT( handle != NULL );
	Node* pNode = handle->GetOwner();
	BOOST_ASSERT( pNode != NULL );
	BOOST_ASSERT( pNode->IsLeaf() );

	Accuracy::Exact distanceSquared(    // twice the squared diameter of the bounding box
		Accuracy::ExactCast( 2.0 ) *
		Accuracy::ExactCast( static_cast< double >( t_Dimension ) )
	);

	++m_tag[ thread ];
	if ( m_tag[ thread ] == 0 ) {

		m_pRoot->ClearTag( thread );
		++m_tag[ thread ];
	}

	std::priority_queue< QueueElement< Node* > > queue;
	typename std::vector< RefcountedPointer< Point > > pSteinerPoints;

	pNode->m_tag[ thread ] = m_tag[ thread ];

	// pNode is a reader
	if ( pNodeList != NULL ) {

		typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeListLock( pNodeList->m_mutex );
		typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );

		pNodeList->m_pNodes.insert( pNode );
		pNode->m_pNodeLists.insert( pNodeList );
	}

	{	typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pNode->m_pPoints.begin();
		typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pNode->m_pPoints.end();
		for ( ; ii != iiEnd; ++ii ) {

			RefcountedPointer< Point > const pPoint = *ii;
			if ( ( pPoint != handle ) && functor( pPoint.Get() ) ) {

				Accuracy::Exact const newDistanceSquared = ExactEuclideanDistanceSquared( *handle, *pPoint );
				if ( newDistanceSquared < distanceSquared ) {

					result = pPoint.Get();
					distanceSquared = newDistanceSquared;
				}
			}
		}
	}

	{	pSteinerPoints.clear();
		{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );
			pSteinerPoints.insert( pSteinerPoints.begin(), pNode->m_pSteinerPoints.begin(), pNode->m_pSteinerPoints.end() );
		}

		typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pSteinerPoints.begin();
		typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pSteinerPoints.end();
		for ( ; ii != iiEnd; ++ii ) {

			RefcountedPointer< Point > const pPoint = *ii;
			if ( ( pPoint != handle ) && functor( pPoint.Get() ) ) {

				Accuracy::Exact const newDistanceSquared = ExactEuclideanDistanceSquared( *handle, *pPoint );
				if ( newDistanceSquared < distanceSquared ) {

					result = pPoint.Get();
					distanceSquared = newDistanceSquared;
				}
			}
		}

		pSteinerPoints.clear();
	}

	// initialize the queue with the neighbors of the current node
	for ( unsigned int ii = 0; ii < ( Power< 3, t_Dimension >::RESULT ); ++ii ) {

		if ( ii != ( Power< 3, t_Dimension >::RESULT ) / 2 ) {

			Node* pNeighbor = pNode->GetNeighbor( ii );
			if ( pNeighbor == NULL ) {

				pNeighbor = pNode->m_pParentNeighbors[ ii ];
				BOOST_ASSERT( ( pNeighbor == NULL ) || ( pNeighbor != pNode->GetParent() ) );
			}
			if ( ( pNeighbor != NULL ) && ( pNeighbor->m_tag[ thread ] < m_tag[ thread ] ) ) {

				Accuracy::Approximate const newApproximateDistanceSquared = ApproximateEuclideanBoxDistanceSquared( pNeighbor->m_lower, pNeighbor->m_upper, *handle );
				queue.push( QueueElement< Node* >( newApproximateDistanceSquared, pNeighbor ) );

				pNeighbor->m_tag[ thread ] = m_tag[ thread ];
			}
		}
	}

	while ( ! queue.empty() ) {

		QueueElement< Node* > const top = queue.top();
		queue.pop();

		if ( Accuracy::Approximate::Compare( top.GetDistance(), distanceSquared.GetApproximate() ) <= 0 ) {

			pNode = top.GetValue();

			if ( pNode->IsLeaf() ) {

				// pNode is a reader
				if ( pNodeList != NULL ) {

					typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeListLock( pNodeList->m_mutex );
					typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );

					pNodeList->m_pNodes.insert( pNode );
					pNode->m_pNodeLists.insert( pNodeList );
				}

				{	typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pNode->m_pPoints.begin();
					typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pNode->m_pPoints.end();
					for ( ; ii != iiEnd; ++ii ) {

						RefcountedPointer< Point > const pPoint = *ii;
						if ( functor( pPoint.Get() ) ) {

							Accuracy::Exact newDistanceSquared = ExactEuclideanDistanceSquared( *handle, *pPoint );
							if ( newDistanceSquared < distanceSquared ) {

								result = pPoint.Get();
								distanceSquared = newDistanceSquared;
							}
						}
					}
				}

				{	pSteinerPoints.clear();
					{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );
						pSteinerPoints.insert( pSteinerPoints.begin(), pNode->m_pSteinerPoints.begin(), pNode->m_pSteinerPoints.end() );
					}

					typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pSteinerPoints.begin();
					typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pSteinerPoints.end();
					for ( ; ii != iiEnd; ++ii ) {

						RefcountedPointer< Point > const pPoint = *ii;
						if ( functor( pPoint.Get() ) ) {

							Accuracy::Exact newDistanceSquared = ExactEuclideanDistanceSquared( *handle, *pPoint );
							if ( newDistanceSquared < distanceSquared ) {

								result = pPoint.Get();
								distanceSquared = newDistanceSquared;
							}
						}
					}

					pSteinerPoints.clear();
				}
			}
			else {

				// enqueue children
				for ( unsigned int ii = 0; ii < ( 1ull << t_Dimension ); ++ii ) {

					BOOST_ASSERT( pNode->m_pChildren != NULL );
					Node* const pChild = &( *pNode->m_pChildren )[ ii ];

					if ( pChild->m_tag[ thread ] < m_tag[ thread ] ) {

						Accuracy::Approximate const newApproximateDistanceSquared = ApproximateEuclideanBoxDistanceSquared( pChild->m_lower, pChild->m_upper, *handle );
						queue.push( QueueElement< Node* >( newApproximateDistanceSquared, pChild ) );

						pChild->m_tag[ thread ] = m_tag[ thread ];
					}
				}
			}

			// enqueue neighbors
			for ( unsigned int ii = 0; ii < ( Power< 3, t_Dimension >::RESULT ); ++ii ) {

				if ( ii != ( Power< 3, t_Dimension >::RESULT ) / 2 ) {

					Node* pNeighbor = pNode->GetNeighbor( ii );
					if ( pNeighbor == NULL ) {

						pNeighbor = pNode->m_pParentNeighbors[ ii ];
						BOOST_ASSERT( ( pNeighbor == NULL ) || ( pNeighbor != pNode->GetParent() ) );
					}
					if ( ( pNeighbor != NULL ) && ( pNeighbor->m_tag[ thread ] < m_tag[ thread ] ) ) {

						Accuracy::Approximate const newApproximateDistanceSquared = ApproximateEuclideanBoxDistanceSquared( pNeighbor->m_lower, pNeighbor->m_upper, *handle );
						queue.push( QueueElement< Node* >( newApproximateDistanceSquared, pNeighbor ) );

						pNeighbor->m_tag[ thread ] = m_tag[ thread ];
					}
				}
			}
		}
	}

	return result;
}


template< unsigned int t_Dimension, unsigned int t_Threads, typename t_Type, typename t_ClientDataType >
template< typename t_Functor >
boost::shared_ptr< ClippedVoronoi< t_Dimension, t_Type, typename Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::Handle > > const Quadtree< t_Dimension, t_Threads, t_Type, t_ClientDataType >::FindClippedVoronoi(
	PoolList& poolList,
	NodeList* const pNodeList,
	Handle const& handle,
	Accuracy::Exact const& nearestNeighborDistanceSquared,
	double const rho,
	double const beta,
	unsigned int const thread,
	t_Functor& functor
) const
{
	BOOST_ASSERT( thread < t_Threads );

	typedef ClippedVoronoi< t_Dimension, t_Type, Handle > CV;
	boost::shared_ptr< CV > pClippedVoronoi( new CV( poolList, *handle, nearestNeighborDistanceSquared, rho, beta ) );

	BOOST_ASSERT( handle != NULL );
	Node* pNode = handle->GetOwner();
	BOOST_ASSERT( pNode != NULL );
	BOOST_ASSERT( pNode->IsLeaf() );

	++m_tag[ thread ];
	if ( m_tag[ thread ] == 0 ) {

		m_pRoot->ClearTag( thread );
		++m_tag[ thread ];
	}

	std::priority_queue< QueueElement< Node* > > nodeQueue;
	std::priority_queue< QueueElement< RefcountedPointer< Point > > > pointQueue;
	typename std::vector< RefcountedPointer< Point > > pSteinerPoints;

	pNode->m_tag[ thread ] = m_tag[ thread ];

	// pNode is a reader
	if ( pNodeList != NULL ) {

		typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeListLock( pNodeList->m_mutex );
		typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );

		pNodeList->m_pNodes.insert( pNode );
		pNode->m_pNodeLists.insert( pNodeList );
	}

	{	typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pNode->m_pPoints.begin();
		typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pNode->m_pPoints.end();
		for ( ; ii != iiEnd; ++ii ) {

			RefcountedPointer< Point > const pPoint = *ii;
			if ( ( pPoint != handle ) && functor( pPoint.Get() ) ) {

				Accuracy::Approximate const newApproximateDistanceSquared = ApproximateEuclideanDistanceSquared( *pPoint, *handle );
				pointQueue.push( QueueElement< RefcountedPointer< Point > >( newApproximateDistanceSquared, pPoint ) );
			}
		}
	}

	{	pSteinerPoints.clear();
		{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );
			pSteinerPoints.insert( pSteinerPoints.begin(), pNode->m_pSteinerPoints.begin(), pNode->m_pSteinerPoints.end() );
		}

		typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pSteinerPoints.begin();
		typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pSteinerPoints.end();
		for ( ; ii != iiEnd; ++ii ) {

			RefcountedPointer< Point > const pPoint = *ii;
			if ( ( pPoint != handle ) && functor( pPoint.Get() ) ) {

				Accuracy::Approximate const newApproximateDistanceSquared = ApproximateEuclideanDistanceSquared( *pPoint, *handle );
				pointQueue.push( QueueElement< RefcountedPointer< Point > >( newApproximateDistanceSquared, pPoint ) );
			}
		}

		pSteinerPoints.clear();
	}

	// initialize the queue with the neighbors of the current node
	for ( unsigned int ii = 0; ii < ( Power< 3, t_Dimension >::RESULT ); ++ii ) {

		if ( ii != ( Power< 3, t_Dimension >::RESULT ) / 2 ) {

			Node* pNeighbor = pNode->GetNeighbor( ii );
			if ( pNeighbor == NULL ) {

				pNeighbor = pNode->m_pParentNeighbors[ ii ];
				BOOST_ASSERT( ( pNeighbor == NULL ) || ( pNeighbor != pNode->GetParent() ) );
			}
			if ( ( pNeighbor != NULL ) && ( pNeighbor->m_tag[ thread ] < m_tag[ thread ] ) ) {

				Accuracy::Approximate const newApproximateDistanceSquared = ApproximateEuclideanBoxDistanceSquared( pNeighbor->m_lower, pNeighbor->m_upper, *handle );
				nodeQueue.push( QueueElement< Node* >( newApproximateDistanceSquared, pNeighbor ) );

				pNeighbor->m_tag[ thread ] = m_tag[ thread ];
			}
		}
	}

	while ( ! nodeQueue.empty() ) {

		QueueElement< Node* > const nodeTop = nodeQueue.top();
		nodeQueue.pop();

		// handle all points which are closer than this node
		while ( ! pointQueue.empty() ) {

			QueueElement< RefcountedPointer< Point > > const pointTop = pointQueue.top();
			if ( Accuracy::Approximate::Compare( pointTop.GetDistance(), nodeTop.GetDistance() ) > 0 )
				break;
			pointQueue.pop();

			if ( pClippedVoronoi->ApproximateTestDistanceSquared( pointTop.GetDistance() ) != CV::TEST_OUTSIDE ) {

				RefcountedPointer< Point > const& pPoint = pointTop.GetValue();

				ClientData< Handle >* const pClientData = pClippedVoronoi->InsertNeighbor( *pPoint );
				if ( pClientData != NULL )
					pClientData->Set( pPoint.Get() );
			}
		}

		pNode = nodeTop.GetValue();
		typename CV::Test const test = pClippedVoronoi->ApproximateTestDistanceSquared( nodeTop.GetDistance() );

		if (
			( test == CV::TEST_INSIDE ) ||
			(
				( test != CV::TEST_OUTSIDE ) &&
				( pClippedVoronoi->ApproximateTestBox( pNode->m_lower, pNode->m_upper ) != CV::TEST_OUTSIDE )
			)
		)
		{
			if ( pNode->IsLeaf() ) {

				// pNode is a reader
				if ( pNodeList != NULL ) {

					typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeListLock( pNodeList->m_mutex );
					typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );

					pNodeList->m_pNodes.insert( pNode );
					pNode->m_pNodeLists.insert( pNodeList );
				}

				{	typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pNode->m_pPoints.begin();
					typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pNode->m_pPoints.end();
					for ( ; ii != iiEnd; ++ii ) {

						RefcountedPointer< Point > const pPoint = *ii;
						if ( functor( pPoint.Get() ) ) {

							Accuracy::Approximate const newApproximateDistanceSquared = ApproximateEuclideanDistanceSquared( *pPoint, *handle );
							pointQueue.push( QueueElement< RefcountedPointer< Point > >( newApproximateDistanceSquared, pPoint ) );
						}
					}
				}

				{	pSteinerPoints.clear();
					{	typename MaybeThreaded< ( t_Threads > 1 ) >::Lock nodeLock( pNode->m_mutex );
						pSteinerPoints.insert( pSteinerPoints.begin(), pNode->m_pSteinerPoints.begin(), pNode->m_pSteinerPoints.end() );
					}

					typename std::vector< RefcountedPointer< Point > >::const_iterator ii    = pSteinerPoints.begin();
					typename std::vector< RefcountedPointer< Point > >::const_iterator iiEnd = pSteinerPoints.end();
					for ( ; ii != iiEnd; ++ii ) {

						RefcountedPointer< Point > const pPoint = *ii;
						if ( functor( pPoint.Get() ) ) {

							Accuracy::Approximate const newApproximateDistanceSquared = ApproximateEuclideanDistanceSquared( *pPoint, *handle );
							pointQueue.push( QueueElement< RefcountedPointer< Point > >( newApproximateDistanceSquared, pPoint ) );
						}
					}

					pSteinerPoints.clear();
				}
			}
			else {

				// enqueue children
				for ( unsigned int ii = 0; ii < ( 1ull << t_Dimension ); ++ii ) {

					BOOST_ASSERT( pNode->m_pChildren != NULL );
					Node* const pChild = &( *pNode->m_pChildren )[ ii ];

					if ( pChild->m_tag[ thread ] < m_tag[ thread ] ) {

						Accuracy::Approximate const newApproximateDistanceSquared = ApproximateEuclideanBoxDistanceSquared( pChild->m_lower, pChild->m_upper, *handle );
						nodeQueue.push( QueueElement< Node* >( newApproximateDistanceSquared, pChild ) );

						pChild->m_tag[ thread ] = m_tag[ thread ];
					}
				}
			}

			// enqueue neighbors
			for ( unsigned int ii = 0; ii < ( Power< 3, t_Dimension >::RESULT ); ++ii ) {

				if ( ii != ( Power< 3, t_Dimension >::RESULT ) / 2 ) {

					Node* pNeighbor = pNode->GetNeighbor( ii );
					if ( pNeighbor == NULL ) {

						pNeighbor = pNode->m_pParentNeighbors[ ii ];
						BOOST_ASSERT( ( pNeighbor == NULL ) || ( pNeighbor != pNode->GetParent() ) );
					}
					if ( ( pNeighbor != NULL ) && ( pNeighbor->m_tag[ thread ] < m_tag[ thread ] ) ) {

						Accuracy::Approximate const newApproximateDistanceSquared = ApproximateEuclideanBoxDistanceSquared( pNeighbor->m_lower, pNeighbor->m_upper, *handle );
						nodeQueue.push( QueueElement< Node* >( newApproximateDistanceSquared, pNeighbor ) );

						pNeighbor->m_tag[ thread ] = m_tag[ thread ];
					}
				}
			}
		}
	}

	// handle any remaining points
	while ( ! pointQueue.empty() ) {

		QueueElement< RefcountedPointer< Point > > const pointTop = pointQueue.top();
		pointQueue.pop();

		if ( pClippedVoronoi->ApproximateTestDistanceSquared( pointTop.GetDistance() ) != CV::TEST_OUTSIDE ) {

			RefcountedPointer< Point > const pPoint = pointTop.GetValue();

			ClientData< Handle >* const pClientData = pClippedVoronoi->InsertNeighbor( *pPoint );
			if ( pClientData != NULL )
				pClientData->Set( pPoint.Get() );
		}
	}

	return pClippedVoronoi;
}




#endif    /* __cplusplus */

#endif    /* __QUADTREE_HH__ */
