/*
	Copyright (C) 2011  Andrew Cotter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
	\internal
	\file vector.hh
	\brief Includes vector headers
*/




#ifndef __VECTOR_HH__
#define __VECTOR_HH__

#ifdef __cplusplus




#include "accuracy.hh"
#include "helpers.hh"

#include <boost/array.hpp>
#include <boost/static_assert.hpp>

#include <cstddef>




namespace {




//============================================================================
//    EuclideanNormSquaredHelper class
//============================================================================


template< typename t_Type, unsigned int t_Dimension, unsigned int t_Index = 1 >
struct EuclideanNormSquaredHelper {

	BOOST_STATIC_ASSERT( t_Index >= 0 );
	BOOST_STATIC_ASSERT( t_Index < t_Dimension );

	typedef typeof(
		Square( ( *static_cast< boost::array< t_Type, t_Dimension >* >( NULL ) )[ t_Index - 1 ] ) +
		*static_cast< typename EuclideanNormSquaredHelper< t_Type, t_Dimension, t_Index + 1 >::ResultType* >( NULL )
	) ResultType;

	static inline ResultType Find( boost::array< t_Type, t_Dimension > const& vector ) {

		return(
			Square( vector[ t_Index - 1 ] ) +
			EuclideanNormSquaredHelper< t_Type, t_Dimension, t_Index + 1 >::Find( vector )
		);
	}
};


template< typename t_Type, unsigned int t_Dimension >
struct EuclideanNormSquaredHelper< t_Type, t_Dimension, t_Dimension > {

	typedef typeof( Square( ( *static_cast< boost::array< t_Type, t_Dimension >* >( NULL ) )[ t_Dimension - 1 ] ) ) ResultType;

	static inline ResultType Find( boost::array< t_Type, t_Dimension > const& vector ) {

		return Square( vector[ t_Dimension - 1 ] );
	}
};




//============================================================================
//    ApproximateEuclideanNormSquaredHelper class
//============================================================================


template< typename t_Type, unsigned int t_Dimension, unsigned int t_Index = 1 >
struct ApproximateEuclideanNormSquaredHelper {

	BOOST_STATIC_ASSERT( t_Index >= 0 );
	BOOST_STATIC_ASSERT( t_Index < t_Dimension );

	typedef Accuracy::Approximate ResultType;

	static inline ResultType Find( boost::array< t_Type, t_Dimension > const& vector ) {

		return(
			Square( ResultType( vector[ t_Index - 1 ] ) ) +
			ApproximateEuclideanNormSquaredHelper< t_Type, t_Dimension, t_Index + 1 >::Find( vector )
		);
	}
};


template< typename t_Type, unsigned int t_Dimension >
struct ApproximateEuclideanNormSquaredHelper< t_Type, t_Dimension, t_Dimension > {

	typedef Accuracy::Approximate ResultType;

	static inline ResultType Find( boost::array< t_Type, t_Dimension > const& vector ) {

		return Square( ResultType( vector[ t_Dimension - 1 ] ) );
	}
};




//============================================================================
//    ExactEuclideanNormSquaredHelper class
//============================================================================


template< typename t_Type, unsigned int t_Dimension, unsigned int t_Index = 1 >
struct ExactEuclideanNormSquaredHelper {

	BOOST_STATIC_ASSERT( t_Index >= 0 );
	BOOST_STATIC_ASSERT( t_Index < t_Dimension );

	typedef typeof(
		Square( Accuracy::ExactCast( ( *static_cast< boost::array< t_Type, t_Dimension >* >( NULL ) )[ t_Index - 1 ] ) ) +
		*static_cast< typename ExactEuclideanNormSquaredHelper< t_Type, t_Dimension, t_Index + 1 >::ResultType* >( NULL )
	) ResultType;

	static inline ResultType Find( boost::array< t_Type, t_Dimension > const& vector ) {

		return(
			Square( Accuracy::ExactCast( vector[ t_Index - 1 ] ) ) +
			ExactEuclideanNormSquaredHelper< t_Type, t_Dimension, t_Index + 1 >::Find( vector )
		);
	}
};


template< typename t_Type, unsigned int t_Dimension >
struct ExactEuclideanNormSquaredHelper< t_Type, t_Dimension, t_Dimension > {

	typedef typeof( Square( Accuracy::ExactCast( ( *static_cast< boost::array< t_Type, t_Dimension >* >( NULL ) )[ t_Dimension - 1 ] ) ) ) ResultType;

	static inline ResultType Find( boost::array< t_Type, t_Dimension > const& vector ) {

		return Square( Accuracy::ExactCast( vector[ t_Dimension - 1 ] ) );
	}
};




//============================================================================
//    EuclideanDistanceSquaredHelper class
//============================================================================


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension, unsigned int t_Index = 1 >
struct EuclideanDistanceSquaredHelper {

	BOOST_STATIC_ASSERT( t_Index >= 0 );
	BOOST_STATIC_ASSERT( t_Index < t_Dimension );

	typedef typeof(
		Square(
			( *static_cast< boost::array< t_LeftType,  t_Dimension >* >( NULL ) )[ t_Index - 1 ] -
			( *static_cast< boost::array< t_RightType, t_Dimension >* >( NULL ) )[ t_Index - 1 ]
		) +
		*static_cast< typename EuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension, t_Index + 1 >::ResultType* >( NULL )
	) ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return(
			Square( left[ t_Index - 1 ] - right[ t_Index - 1 ] ) +
			EuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension, t_Index + 1 >::Find( left, right )
		);
	}
};


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension >
struct EuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension, t_Dimension > {

	typedef typeof(
		Square(
			( *static_cast< boost::array< t_LeftType,  t_Dimension >* >( NULL ) )[ t_Dimension - 1 ] -
			( *static_cast< boost::array< t_RightType, t_Dimension >* >( NULL ) )[ t_Dimension - 1 ]
		)
	) ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return Square( left[ t_Dimension - 1 ] - right[ t_Dimension - 1 ] );
	}
};




//============================================================================
//    ApproximateEuclideanDistanceSquaredHelper class
//============================================================================


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension, unsigned int t_Index = 1 >
struct ApproximateEuclideanDistanceSquaredHelper {

	BOOST_STATIC_ASSERT( t_Index >= 0 );
	BOOST_STATIC_ASSERT( t_Index < t_Dimension );

	typedef Accuracy::Approximate ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return(
			Square( ResultType( left[ t_Index - 1 ] ) - ResultType( right[ t_Index - 1 ] ) ) +
			ApproximateEuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension, t_Index + 1 >::Find( left, right )
		);
	}
};


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension >
struct ApproximateEuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension, t_Dimension > {

	typedef Accuracy::Approximate ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return Square( ResultType( left[ t_Dimension - 1 ] ) - ResultType( right[ t_Dimension - 1 ] ) );
	}
};




//============================================================================
//    ExactEuclideanDistanceSquaredHelper class
//============================================================================


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension, unsigned int t_Index = 1 >
struct ExactEuclideanDistanceSquaredHelper {

	BOOST_STATIC_ASSERT( t_Index >= 0 );
	BOOST_STATIC_ASSERT( t_Index < t_Dimension );

	typedef typeof(
		Square(
			Accuracy::ExactCast( ( *static_cast< boost::array< t_LeftType,  t_Dimension >* >( NULL ) )[ t_Index - 1 ] ) -
			Accuracy::ExactCast( ( *static_cast< boost::array< t_RightType, t_Dimension >* >( NULL ) )[ t_Index - 1 ] )
		) +
		*static_cast< typename ExactEuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension, t_Index + 1 >::ResultType* >( NULL )
	) ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return(
			Square( Accuracy::ExactCast( left[ t_Index - 1 ] ) - Accuracy::ExactCast( right[ t_Index - 1 ] ) ) +
			ExactEuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension, t_Index + 1 >::Find( left, right )
		);
	}
};


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension >
struct ExactEuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension, t_Dimension > {

	typedef typeof(
		Square(
			Accuracy::ExactCast( ( *static_cast< boost::array< t_LeftType,  t_Dimension >* >( NULL ) )[ t_Dimension - 1 ] ) -
			Accuracy::ExactCast( ( *static_cast< boost::array< t_RightType, t_Dimension >* >( NULL ) )[ t_Dimension - 1 ] )
		)
	) ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return Square( Accuracy::ExactCast( left[ t_Dimension - 1 ] ) - Accuracy::ExactCast( right[ t_Dimension - 1 ] ) );
	}
};




//============================================================================
//    DotProductHelper class
//============================================================================


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension, unsigned int t_Index = 1 >
struct DotProductHelper {

	BOOST_STATIC_ASSERT( t_Index >= 0 );
	BOOST_STATIC_ASSERT( t_Index < t_Dimension );

	typedef typeof(
		( *static_cast< boost::array< t_LeftType,  t_Dimension >* >( NULL ) )[ t_Index - 1 ] *
		( *static_cast< boost::array< t_RightType, t_Dimension >* >( NULL ) )[ t_Index - 1 ] +
		*static_cast< typename DotProductHelper< t_LeftType, t_RightType, t_Dimension, t_Index + 1 >::ResultType* >( NULL )
	) ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return(
			left[ t_Index - 1 ] * right[ t_Index - 1 ] +
			DotProductHelper< t_LeftType, t_RightType, t_Dimension, t_Index + 1 >::Find( left, right )
		);
	}
};


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension >
struct DotProductHelper< t_LeftType, t_RightType, t_Dimension, t_Dimension > {

	typedef typeof(
		( *static_cast< boost::array< t_LeftType,  t_Dimension >* >( NULL ) )[ t_Dimension - 1 ] *
		( *static_cast< boost::array< t_RightType, t_Dimension >* >( NULL ) )[ t_Dimension - 1 ]
	) ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return( left[ t_Dimension - 1 ] * right[ t_Dimension - 1 ] );
	}
};




//============================================================================
//    ApproximateDotProductHelper class
//============================================================================


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension, unsigned int t_Index = 1 >
struct ApproximateDotProductHelper {

	BOOST_STATIC_ASSERT( t_Index >= 0 );
	BOOST_STATIC_ASSERT( t_Index < t_Dimension );

	typedef Accuracy::Approximate ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return(
			ResultType( left[ t_Index - 1 ] ) * ResultType( right[ t_Index - 1 ] ) +
			ApproximateDotProductHelper< t_LeftType, t_RightType, t_Dimension, t_Index + 1 >::Find( left, right )
		);
	}
};


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension >
struct ApproximateDotProductHelper< t_LeftType, t_RightType, t_Dimension, t_Dimension > {

	typedef Accuracy::Approximate ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return( ResultType( left[ t_Dimension - 1 ] ) * ResultType( right[ t_Dimension - 1 ] ) );
	}
};




//============================================================================
//    ExactDotProductHelper class
//============================================================================


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension, unsigned int t_Index = 1 >
struct ExactDotProductHelper {

	BOOST_STATIC_ASSERT( t_Index >= 0 );
	BOOST_STATIC_ASSERT( t_Index < t_Dimension );

	typedef typeof(
		Accuracy::ExactCast( ( *static_cast< boost::array< t_LeftType,  t_Dimension >* >( NULL ) )[ t_Index - 1 ] ) *
		Accuracy::ExactCast( ( *static_cast< boost::array< t_RightType, t_Dimension >* >( NULL ) )[ t_Index - 1 ] ) +
		*static_cast< typename ExactDotProductHelper< t_LeftType, t_RightType, t_Dimension, t_Index + 1 >::ResultType* >( NULL )
	) ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return(
			Accuracy::ExactCast( left[ t_Index - 1 ] ) * Accuracy::ExactCast( right[ t_Index - 1 ] ) +
			ExactDotProductHelper< t_LeftType, t_RightType, t_Dimension, t_Index + 1 >::Find( left, right )
		);
	}
};


template< typename t_LeftType, typename t_RightType, unsigned int t_Dimension >
struct ExactDotProductHelper< t_LeftType, t_RightType, t_Dimension, t_Dimension > {

	typedef typeof(
		Accuracy::ExactCast( ( *static_cast< boost::array< t_LeftType,  t_Dimension >* >( NULL ) )[ t_Dimension - 1 ] ) *
		Accuracy::ExactCast( ( *static_cast< boost::array< t_RightType, t_Dimension >* >( NULL ) )[ t_Dimension - 1 ] )
	) ResultType;

	static inline ResultType Find(
		boost::array< t_LeftType,  t_Dimension > const& left,
		boost::array< t_RightType, t_Dimension > const& right
	)
	{
		return( Accuracy::ExactCast( left[ t_Dimension - 1 ] ) * Accuracy::ExactCast( right[ t_Dimension - 1 ] ) );
	}
};




}    // anonymous namespace




//============================================================================
//    vector functions
//============================================================================


template< typename t_Type, std::size_t t_Dimension >
inline typename EuclideanNormSquaredHelper< t_Type, t_Dimension >::ResultType EuclideanNormSquared( boost::array< t_Type, t_Dimension > const& vector ) {

	return EuclideanNormSquaredHelper< t_Type, t_Dimension >::Find( vector );
}


template< typename t_Type, std::size_t t_Dimension >
inline typename ApproximateEuclideanNormSquaredHelper< t_Type, t_Dimension >::ResultType ApproximateEuclideanNormSquared( boost::array< t_Type, t_Dimension > const& vector ) {

	return ApproximateEuclideanNormSquaredHelper< t_Type, t_Dimension >::Find( vector );
}


template< typename t_Type, std::size_t t_Dimension >
inline typename ExactEuclideanNormSquaredHelper< t_Type, t_Dimension >::ResultType ExactEuclideanNormSquared( boost::array< t_Type, t_Dimension > const& vector ) {

	return ExactEuclideanNormSquaredHelper< t_Type, t_Dimension >::Find( vector );
}


template< typename t_LeftType, typename t_RightType, std::size_t t_Dimension >
inline typename EuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension >::ResultType EuclideanDistanceSquared(
	boost::array< t_LeftType,  t_Dimension > const& left,
	boost::array< t_RightType, t_Dimension > const& right
)
{
	return EuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension >::Find( left, right );
}


template< typename t_LeftType, typename t_RightType, std::size_t t_Dimension >
inline typename ApproximateEuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension >::ResultType ApproximateEuclideanDistanceSquared(
	boost::array< t_LeftType,  t_Dimension > const& left,
	boost::array< t_RightType, t_Dimension > const& right
)
{
	return ApproximateEuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension >::Find( left, right );
}


template< typename t_LeftType, typename t_RightType, std::size_t t_Dimension >
inline typename ExactEuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension >::ResultType ExactEuclideanDistanceSquared(
	boost::array< t_LeftType,  t_Dimension > const& left,
	boost::array< t_RightType, t_Dimension > const& right
)
{
	return ExactEuclideanDistanceSquaredHelper< t_LeftType, t_RightType, t_Dimension >::Find( left, right );
}


template< typename t_LeftType, typename t_RightType, std::size_t t_Dimension >
inline typename DotProductHelper< t_LeftType, t_RightType, t_Dimension >::ResultType DotProduct(
	boost::array< t_LeftType,  t_Dimension > const& left,
	boost::array< t_RightType, t_Dimension > const& right
)
{
	return DotProductHelper< t_LeftType, t_RightType, t_Dimension >::Find( left, right );
}


template< typename t_LeftType, typename t_RightType, std::size_t t_Dimension >
inline typename ApproximateDotProductHelper< t_LeftType, t_RightType, t_Dimension >::ResultType ApproximateDotProduct(
	boost::array< t_LeftType,  t_Dimension > const& left,
	boost::array< t_RightType, t_Dimension > const& right
)
{
	return ApproximateDotProductHelper< t_LeftType, t_RightType, t_Dimension >::Find( left, right );
}


template< typename t_LeftType, typename t_RightType, std::size_t t_Dimension >
inline typename ExactDotProductHelper< t_LeftType, t_RightType, t_Dimension >::ResultType ExactDotProduct(
	boost::array< t_LeftType,  t_Dimension > const& left,
	boost::array< t_RightType, t_Dimension > const& right
)
{
	return ExactDotProductHelper< t_LeftType, t_RightType, t_Dimension >::Find( left, right );
}




#endif    /* __cplusplus */

#endif    /* __VECTOR_HH__ */
